"""
The Controller class manages a collection of Components and associates them with
a CustomResource in the cluster.
"""

# Standard
from typing import Optional, Tuple, Union
import abc

# First Party
import aconfig
import alog

# Local
from .dag import CompletionState
from .exceptions import Oper8Error, RolloutError
from .reconcile import RequeueParams
from .rollout_manager import RolloutManager
from .session import Session
from .utils import abstractclassproperty, classproperty
from .verify_resources import verify_subsystem

## Globals #####################################################################

log = alog.use_channel("CTRLR")


## Application #################################################################


class Controller(abc.ABC):
    """This class represents a controller for a single kubernetes custom
    resource kind. Its primary functionality is to perform a reconciliation of a
    given CR manifest for an instance of the resource kind against the current
    state of the cluster. To accomplish this, its reconciliation logic is:

    1. Construct a Directed Acyclic Graph of all Components that this kind
        needs to manage.
    2. Execute the Graph in dependency order where each node of the graph first
        renders the manifests for all kubernetes resources managed by the
        Component, then applies them to the cluster.
    3. Execute a secondary Graph with verification logic for each Component,
        terminating verification for downstream nodes if any node is not yet
        verified.

    To do this, the main operations of the class are to construct a DAG of
    Components, then walk them through the primary lifecycle phases:

    1. Run the Component's deploy() function to completion and verify that the
        actual deployment operations succeeded
    2. Run the Component's verify() function to run component-specific tests
        that will verify if the deployment is rolled out in a successful state
    """

    ## Class Properties ########################################################

    # Derived classes must have class properties for group, version, and kind.
    # To enforce this, we set defaults for all of these and then validate that
    # they are present, we define them as classproperty and raise when accessed
    # from the base implementation.

    # NOTE: pylint is very confused by the use of these property decorators, so
    #   we need to liberally ignore warnings.

    @abstractclassproperty  # noqa: B027
    def group(cls) -> str:
        """The apiVersion group for the resource this controller manages"""

    @abstractclassproperty  # noqa: B027
    def version(cls) -> str:
        """The apiVersion version for the resource this controller manages"""

    @abstractclassproperty  # noqa: B027
    def kind(cls) -> str:
        """The kind for the resource this controller manages"""

    @classproperty
    def finalizer(cls) -> Optional[str]:  # pylint: disable=no-self-argument
        """The finalizer used by this Controller"""
        if cls.has_finalizer:  # pylint: disable=using-constant-test
            return f"finalizers.{cls.kind.lower()}.{cls.group}"  # pylint: disable=no-member
        return None

    @classproperty
    def has_finalizer(cls) -> bool:  # pylint: disable=no-self-argument
        """If the derived class has an implementation of finalize_components, it
        has a finalizer and can be registered for finalize events
        """
        return cls.finalize_components is not Controller.finalize_components

    ## Construction ############################################################

    def __init__(self, config_defaults: Optional[aconfig.Config] = None):
        """The constructor sets up all of the properties of the controller which
        are constant across reconciliations.

        Args:
            config_defaults:  Optional[aconfig.Config]
                Default values for the backend controller config

        """
        # Make sure the class properties are present and not empty
        assert self.group, "Controller.group must be a non-empty string"
        assert self.version, "Controller.version must be a non-empty string"
        assert self.kind, "Controller.kind must be a non-empty string"
        self.config_defaults = config_defaults or aconfig.Config({})

    @classmethod
    def __str__(cls):
        """Stringify with the GVK"""
        return f"Controller({cls.group}/{cls.version}/{cls.kind})"

    ## Abstract Interface ######################################################
    #
    # These functions must be implemented by child classes
    ##

    @abc.abstractmethod
    def setup_components(self, session: Session):
        """Given the session for an individual reconciliation, construct the set
        of Components that will be deployed.

        Error Semantics: Child classes should throw ConfigError if config is
        not valid and include the portion of config where the problem occurred.

        Args:
            session:  Session
                The current session containing the per-event configs
        """

    ## Base Class Interface ####################################################
    #
    # These methods MAY be implemented by children, but contain default
    # implementations that are appropriate for simple cases.
    #
    # NOTE: We liberally use pylint disables here to make the base interface
    #   clear to deriving classes.
    ##

    def finalize_components(self, session: Session):  # noqa: B027
        """When performing a finalizer operation, this function will be called
        to perform custom finalizer logic for this Controller.

        Error Semantics: Child classes should throw ConfigError if config is
        not valid and include the portion of config where the problem occurred.

        NOTE: This method is not abstract since the standard controller usecase
            does not require finalizing

        Args:
            session:  Session
                The current session containing the per-event configs
        """

    def after_deploy(
        self, session: Session, deploy_completion_state: CompletionState
    ) -> bool:
        """This allows children to inject logic that will run when the
        controller has finished deploying all components, but not necessarily
        verifying all of them. The default behavior is a no-op.

        Args:
            session:  Session
                The current reconciliation session
            deploy_completion_state: CompletionState
                Completion state of the last deployment

        Returns:
            success:  bool
                True if custom hook code executed successfully and lifecycle
                should continue
        """
        return True

    def after_deploy_unsuccessful(
        self, session: Session, failed: bool, deploy_completion_state: CompletionState
    ) -> bool:
        """This allows children to inject logic that will run when the
        controller has failed or could not finish deploying all components.
        The default behavior is a no-op.

        Args:
            session:  Session
                The current reconciliation session
            failed:  bool
                Indicator of whether or not the termination was a failure
            deploy_completion_state: CompletionState
                Completion state of the last deployment

        Returns:
            success:  bool
                True if custom hook code executed successfully and lifecycle
                should continue
        """
        return True

    def after_verify(
        self,
        session: Session,  # pylint: disable=unused-argument
        verify_completion_state: CompletionState,
        deploy_completion_state: CompletionState,
    ) -> bool:
        """This allows children to inject logic that will run when the
        controller has finished verifying all components. The default behavior
        is a no-op.

        Args:
            session:  Session
                The current reconciliation session
            verify_completion_state: CompletionState
                Completion state of the last verification
            deploy_completion_state: CompletionState
                Completion state of the last deployment

        Returns:
            success:  bool
                True if custom hook code executed successfully and lifecycle
                should continue
        """
        return True

    def after_verify_unsuccessful(
        self,
        session: Session,
        failed: bool,
        verify_completion_state: CompletionState,
        deploy_completion_state: CompletionState,
    ) -> bool:
        """This allows children to inject logic that will run when the
        controller has finished deploying all components but failed to verify.
        The default behavior is a no-op.

        Args:
            session:  Session
                The current reconciliation session
            failed:  bool
                Indicator of whether or not the termination was a failure
            verify_completion_state: CompletionState
                Completion state of the last verification
            deploy_completion_state: CompletionState
                Completion state of the last deployment

        Returns:
            success:  bool
                True if custom hook code executed successfully and lifecycle
                should continue
        """
        return True

    def should_requeue(self, session: Session) -> Tuple[bool, Optional[RequeueParams]]:
        """should_requeue determines if current reconcile request should be re-queued.

        Children can override default implementation to provide custom logic.
        Default implementation re-queues the request if the reconciling CR status
        hasn't been reached stable state.

        Args:
            session: Session
                The current reconciliation session

        Returns:
            requeue: bool
                True if the reconciliation request should be re-queued
            config: RequeueParams
                 Parameters of requeue request. Can be None if requeue is False.
        """
        api_version = session.api_version
        kind = session.kind
        name = session.name
        namespace = session.namespace
        requeue_params = RequeueParams()
        # Fetch the current status from the cluster
        success, current_state = session.deploy_manager.get_object_current_state(
            api_version=api_version,
            kind=kind,
            name=name,
            namespace=namespace,
        )
        if not success:
            log.warning(
                "Failed to fetch current state for %s/%s/%s", namespace, kind, name
            )
            return True, requeue_params
        # Do not requeue if resource was deleted
        if not current_state:
            log.warning("Resource not found: %s/%s/%s", namespace, kind, name)
            return False, requeue_params

        log.debug3("Current CR manifest for requeue check: %s", current_state)

        verified = verify_subsystem(current_state, session.version)
        return not verified, requeue_params

    def get_cr_manifest_defaults(
        self,
    ) -> Union[dict, aconfig.Config]:
        """This allows children to provide default values for their cr_manifest
        that will be injected where no override is provided in the user-provided
        cr_manifest.

        Returns:
            cr_manifest_defaults:  Union[dict, aconfig.Config]
                The cr defaults. Raw dicts will be converted to Config objects.
        """
        return aconfig.Config({})

    def get_config_defaults(self):
        """This function allows children to override the default values for the session
        config. This value can also be set via the controllers __init__ function.
        """
        return self.config_defaults

    ## Public Interface ########################################################
    #
    # These functions should be used by the reconciliation manager or in
    # tests
    ##

    def run_reconcile(
        self, session: Session, is_finalizer: bool = False
    ) -> CompletionState:
        """Perform a reconciliation iteration for this controller on given a session.
        This function should only be called once per session. The general logic for a
        controller reconcile is as follows:

        1. Set up the set of Components and their dependencies that will be
            managed in this reconciliation based on the CR and config
        2. Invoke the rollout to render each component and apply it to the
            cluster (if not in dry-run), then verify the DAG of components

        Args:
            session:  Session
                The full structured content of the CR manifest for this operand
            is_finalizer:  bool
                If true, the logic in finalize_components is run, otherwise the
                logic in setup_components is called

        Returns:
            result: ReconciliationResult
                The result of reconcile
        """
        # Check if session has already been reconciled
        if not session.graph.empty():
            raise RolloutError("Session has already been reconciled")

        self._manage_components(session, is_finalizer)
        completion_state = self._rollout_components(session)
        return completion_state

    ## Implementation Details ##################################################

    def _manage_components(self, session: Session, is_finalizer: bool):
        """Delegate logic to child's finalize_components or setup_components

        Args:
            session: Session
                The current session being reconciled
            is_finalizer: bool
                Weather the current CR is being deleted

        """

        # If this is a finalizer, run finalize_components
        if is_finalizer:
            log.debug("[%s] Running as finalizer", session.id)
            self.finalize_components(session)

        # Otherwise run setup_components
        else:
            self.setup_components(session)

    @alog.logged_function(log.debug)
    def _rollout_components(self, session: Session):
        """Deploy all dependent components according to the configured
        dependencies between them
        """
        log.debug("Rolling out %s", str(self))
        log.debug3("Session dependency DAG: %s", str(session.graph))

        # Set up the deployment manager and run the rollout
        rollout_manager = RolloutManager(
            session=session,
            after_deploy=self.after_deploy,
            after_deploy_unsuccessful=self.after_deploy_unsuccessful,
            after_verify=self.after_verify,
            after_verify_unsuccessful=self.after_verify_unsuccessful,
        )
        completion_state = rollout_manager.rollout()
        rollout_failed = completion_state.failed()
        log.info("Final rollout state: %s", completion_state)

        # Get Rollout Status
        deploy_completed = completion_state.deploy_completed()
        verify_completed = completion_state.verify_completed()
        log.debug2(
            "Deploy Completed: %s, Verify Completed: %s, Deploy Failed: %s",
            deploy_completed,
            verify_completed,
            rollout_failed,
        )

        # If an oper8 error occurred in the rollout, decorate it with a reference
        # to the completion state itself and then raise it to be handled by the
        # top-level ReconcileManager handler.
        if isinstance(completion_state.exception, Oper8Error):
            log.debug("Handling Oper8Error from rollout")
            completion_state.exception.completion_state = completion_state
            raise completion_state.exception

        # If the deploy failed but didn't trigger an Oper8Error, we'll make one
        # ourselves
        if rollout_failed:
            raise RolloutError(
                "Deploy phase failed", completion_state=completion_state
            ) from completion_state.exception

        return completion_state



"""
CLI command for setting up a VCS version repo
"""
# Standard
import argparse

# First Party
import alog

# Local
from ..setup_vcs import DEFAULT_DEST, DEFAULT_TAG_EXPR, setup_vcs
from .base import CmdBase

log = alog.use_channel("CMD-VCS")


class SetupVCSCmd(CmdBase):
    __doc__ = __doc__

    def add_subparser(
        self,
        subparsers: argparse._SubParsersAction,
    ) -> argparse.ArgumentParser:
        """Add the subparser for this command"""
        parser = subparsers.add_parser(
            "setup-vcs",
            help="Initialize a clean git repo to use with VCS versioning",
        )
        command_args = parser.add_argument_group("Command Arguments")
        command_args.add_argument(
            "--source",
            "-s",
            required=True,
            help="Source repo to seed the clean git history",
        )
        command_args.add_argument(
            "--destination",
            "-d",
            default=DEFAULT_DEST,
            help="Destination directory in which to place the clean git history",
        )
        command_args.add_argument(
            "--branch-expr",
            "-b",
            nargs="*",
            default=None,
            help="Regular expression(s) to use to identify branches",
        )
        command_args.add_argument(
            "--tag-expr",
            "-te",
            nargs="*",
            default=DEFAULT_TAG_EXPR,
            help="Regular expression(s) to use to identify tags",
        )
        command_args.add_argument(
            "--force",
            "-f",
            action="store_true",
            default=False,
            help="Force overwrite existing destination",
        )
        return parser

    def cmd(self, args: argparse.Namespace):
        setup_vcs(
            source=args.source,
            destination=args.destination,
            branch_expr=args.branch_expr,
            tag_expr=args.tag_expr,
            force=args.force,
        )



"""
Base class interface for a Postgres component
"""

# Local
from .... import component
from ..interfaces import Datastore

COMPONENT_NAME = "postgres"


@component(COMPONENT_NAME)
class IPostgresComponent(Datastore):
    """A postgres chart provides access to a single running Postgres cluster"""

    ## Shared Utilities ########################################################

    def tls_enabled(self) -> bool:
        """Return whether TLS is enabled or not
        Returns:
            bool: True (TLS enabled), False (TLS disabled)
        """
        return self.config.get("tls", {}).get("enabled", True)



"""
Tests of the redis factory
"""
# Third Party
import pytest

# Local
from oper8.test_helpers.helpers import setup_session
from oper8.x.datastores.redis.factory import RedisFactory

## Helpers #####################################################################


def get_config_overrides(config):
    return {RedisFactory.DATASTORE_TYPE: config}


## Error cases #################################################################


def test_construct_unknown():
    """ " Test error when unsupport type to RedisFactory"""
    session = setup_session(app_config=get_config_overrides({"type": "UnknownType"}))
    with pytest.raises(AssertionError):
        RedisFactory.get_component(session)


def test_construct_notype():
    """ " Test error when no type specified to RedisFactory"""
    session = setup_session(app_config=get_config_overrides({"type": "ToRemove"}))
    if session.config.redis["type"] is not None:
        del session.config.redis["type"]
    with pytest.raises(AssertionError):
        RedisFactory.get_component(session)



"""
Shared module to hold constant values for the library
"""

# Reconciliation configuration annotations
PAUSE_ANNOTATION_NAME = "oper8.org/pause-execution"
CONFIG_DEFAULTS_ANNOTATION_NAME = "oper8.org/config-defaults"

# Leadership annotations
LEASE_NAME_ANNOTATION_NAME = "oper8.org/lease-name"
LEASE_TIME_ANNOTATION_NAME = "oper8.org/lease-time"

# Log config annotations
LOG_DEFAULT_LEVEL_NAME = "oper8.org/log-default-level"
LOG_FILTERS_NAME = "oper8.org/log-filters"
LOG_THREAD_ID_NAME = "oper8.org/log-thread-id"
LOG_JSON_NAME = "oper8.org/log-json"

# List to keep track of all oper8-managed annotations which should be passed
# from a parent Application to a child CR.
# NOTE: The only excluded annotations are
# * temporary patches as this is managed by the oper8_temporary_patch module
#   directly
# * leadership annotations since those may differ per CR instance
PASSTHROUGH_ANNOTATIONS = [
    CONFIG_DEFAULTS_ANNOTATION_NAME,
    LOG_DEFAULT_LEVEL_NAME,
    LOG_FILTERS_NAME,
    LOG_JSON_NAME,
    LOG_THREAD_ID_NAME,
    PAUSE_ANNOTATION_NAME,
]

# BACKWARDS COMPATIBILITY: We maintain the ALL_ANNOTATIONS name for
# compatibility with old code that accessed this directly
ALL_ANNOTATIONS = PASSTHROUGH_ANNOTATIONS

# The name of the annotation used to attach TemporaryPatch resources to a given
# oper8-managed CR
TEMPORARY_PATCHES_ANNOTATION_NAME = "oper8.org/temporary-patches"

# The name of the annotation used to indicate the internal name of each
# oper8-managed resource
INTERNAL_NAME_ANNOTATION_NAME = "oper8.org/internal-name"
# Keeping the misspelled variable for backward compatibility: https://github.com/IBM/oper8/pull/133#discussion_r1820696481
INTERNAL_NAME_ANOTATION_NAME = INTERNAL_NAME_ANNOTATION_NAME

# Default namespace if none given
DEFAULT_NAMESPACE = "default"

# Delimiter used for nested dict keys
NESTED_DICT_DELIM = "."

# Name of the spec section used to provide config overrides
CONFIG_OVERRIDES = "configOverrides"



"""
Tests for the DryRunWatchManager
"""

# Third Party
import pytest

# Local
from oper8 import Controller, controller
from oper8.test_helpers.helpers import (
    DummyController,
    MockDeployManager,
    library_config,
    setup_cr,
)
from oper8.watch_manager import DryRunWatchManager, WatchManagerBase

## Helpers #####################################################################


class FooController(DummyController):
    def __init__(self, **kwargs):
        super().__init__(
            components=[
                {
                    "name": "foo",
                    "api_objects": [("foo", {"kind": "Foo", "apiVersion": "v1"})],
                }
            ],
            **kwargs,
        )


class BarController(FooController):
    """Testable controller with a finalizer"""

    kind = "Bar"

    FINALIZE_CALLED = False

    def finalize_components(self, session):
        self.__class__.FINALIZE_CALLED = True


@pytest.fixture(autouse=True)
def reset_globals():
    """This helper is only used in tests to "reset" the state of the global
    watches dict
    """
    WatchManagerBase._ALL_WATCHES = {}
    BarController.FINALIZE_CALLED = False


## Tests #######################################################################


def test_lazy_construct():
    """Make sure that the controller is not constructed when the watch manager
    is first instantiated
    """

    @controller(group="foo.bar", version="v1", kind="Foo")
    class BadController(Controller):
        def __init__(self, *_, **__):
            raise RuntimeError()

    DryRunWatchManager(BadController)


def test_double_watch():
    """Test that all constructed watch managers get registered"""
    wm = DryRunWatchManager(FooController)
    assert wm.watch()
    assert not wm.watch()


def test_reconcile_single():
    """Test that a single watched controller can receive CR reconcile events"""
    dm = MockDeployManager()
    cr = setup_cr(
        kind=FooController.kind,
        api_version=f"{FooController.group}/{FooController.version}",
    )
    DryRunWatchManager(FooController, deploy_manager=dm)
    DryRunWatchManager.start_all()
    dm.deploy([cr])
    assert dm.has_obj(
        kind="Foo", api_version="v1", name="foo", namespace=cr.metadata.namespace
    )


def test_finalize_called():
    """Test that when an object exists in the cluster and is deleted, the
    finalizer for a Controller with one defined will be called
    """
    cr = setup_cr(
        kind=BarController.kind,
        api_version=f"{BarController.group}/{BarController.version}",
    )
    dm = MockDeployManager(resources=[cr])
    DryRunWatchManager(BarController, deploy_manager=dm)
    DryRunWatchManager.start_all()
    assert dm.has_obj(
        kind=cr.kind,
        api_version=cr.apiVersion,
        name=cr.metadata.name,
        namespace=cr.metadata.namespace,
    )
    dm.disable([cr])
    assert not dm.has_obj(
        kind=cr.kind,
        api_version=cr.apiVersion,
        name=cr.metadata.name,
        namespace=cr.metadata.namespace,
    )
    assert BarController.FINALIZE_CALLED


def test_watch_subsystem():
    """Test that a CR generated by one controller can trigger another to run"""

    # CR for the parent
    cr = setup_cr(kind="Parent", api_version="foo.bar/v1")

    # Subsystem controller class that createsd a Foo api object
    @controller(group="foo.bar", version="v1", kind="Subsystem")
    class SubsystemController(DummyController):
        def __init__(self):
            super().__init__(
                components=[
                    {
                        "name": "foo",
                        "api_objects": [("foo", {"kind": "Foo", "apiVersion": "v1"})],
                    }
                ],
            )

    # Parent controller that creates a Bar api object and the CR for the
    # subsystem
    @controller(group="foo.bar", version="v1", kind="Parent")
    class ParentController(DummyController):
        def __init__(self):
            super().__init__(
                components=[
                    {
                        "name": "foo-subsystem",
                        "api_objects": [
                            (
                                cr.metadata.name,
                                {"kind": "Subsystem", "apiVersion": "foo.bar/v1"},
                            )
                        ],
                    },
                    {
                        "name": "bar",
                        "api_objects": [("bar", {"kind": "Bar", "apiVersion": "v1"})],
                    },
                ],
            )

    # Set up both controllers to watch
    dm = MockDeployManager()
    DryRunWatchManager(ParentController, deploy_manager=dm)
    DryRunWatchManager(SubsystemController, deploy_manager=dm)
    DryRunWatchManager.start_all()

    # Deploy the CR
    with library_config(standalone=True):
        dm.deploy([cr])

    # Make sure the resources for both the parent and subsystem got deployed
    assert dm.has_obj(
        kind="Foo", name="foo", api_version="v1", namespace=cr.metadata.namespace
    )
    assert dm.has_obj(
        kind="Bar", name="bar", api_version="v1", namespace=cr.metadata.namespace
    )



"""Tests for the DryRunDeployManager

NOTE: The majority of the functionality is thoroughly exercised by all of the
    other unit tests, so the tests here only test elements that are particularly
    delicate and/or not covered elsewhere.
"""
# Standard
from threading import Timer
from unittest.mock import Mock

# Third Party
import pytest

# First Party
import aconfig

# Local
from oper8.deploy_manager import DryRunDeployManager, KubeEventType, KubeWatchEvent
from oper8.deploy_manager.owner_references import _make_owner_reference
from oper8.test_helpers.helpers import SOME_OTHER_NAMESPACE, TEST_NAMESPACE, setup_cr

## Helpers #####################################################################


def make_obj(
    api_version, kind, name="foobar", namespace=SOME_OTHER_NAMESPACE, spec={"a": 1}
):
    return aconfig.Config(
        {
            "apiVersion": api_version,
            "kind": kind,
            "metadata": {
                "name": name,
                "namespace": namespace,
                "labels": {
                    "app": "foobar",
                    "run": "frontend",
                },
            },
            "spec": spec,
        },
        override_env_vars=False,
    )


## Tests #######################################################################


def test_watches_triggered():
    """Test that registered watches are triggered when a resource of the right
    GVK is deployed
    """
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    mock1 = Mock()
    mock2 = Mock()
    dm.register_watch(api_version, kind, mock1)
    dm.register_watch(api_version, kind, mock2)
    dm.deploy([obj])

    mock1.assert_called_once_with(obj)
    mock2.assert_called_once_with(obj)


def test_watches_not_triggered_other_resource():
    """Test that a registered watch is not triggered for an object with a
    different GVK
    """
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    mock = Mock()
    dm.register_watch("foo.bar/v2", kind, mock)
    dm.deploy([obj])
    assert not mock.called


def test_finalizer_triggered():
    """Test that registered finalizers are triggered when a resource of the
    right GVK is deleted
    """
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    mock1 = Mock()
    mock2 = Mock()
    dm.register_finalizer(api_version, kind, mock1)
    dm.register_finalizer(api_version, kind, mock2)
    dm.deploy([obj])
    assert not mock1.called
    assert not mock2.called
    dm.disable([obj])
    assert mock1.call_count == 1
    assert mock2.call_count == 1


def test_finalizer_not_triggered_not_deleted():
    """Test that a registered finalizer is not triggered when a resource of the
    right GVK is disabled, but was not present in the cluster
    """
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    mock = Mock()
    dm.register_finalizer(api_version, kind, mock)
    dm.disable([obj])
    assert not mock.called


def test_finalizer_not_triggered_other_resource():
    """Test that a registered finalizer is not triggered when a resource of the
    right GVK is disabled, but was not present in the cluster
    """
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    mock = Mock()
    dm.register_finalizer("foo.bar/v2", kind, mock)
    dm.deploy([obj])
    assert not mock.called
    dm.disable([obj])
    assert not mock.called


def test_watch_key_none():
    """Make sure the _watch_key function can handle None arguments. This is used
    at times in unit tests and should be supported
    """
    dm = DryRunDeployManager()
    assert isinstance(dm._watch_key(None, None), str)


@pytest.mark.timeout(5)
def test_watch_objects():
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    dm.deploy([obj])

    deploy_timer_event = Timer(
        interval=0.5,
        function=dm.deploy,
        args=[[make_obj(api_version, kind, name="barfoo")]],
    )
    update_timer_event = Timer(
        interval=1,
        function=dm.deploy,
        args=[[make_obj(api_version, kind, name="barfoo", spec={"modified": 1})]],
    )
    disable_timer_event = Timer(
        interval=1.5, function=dm.disable, args=[[make_obj(api_version, kind)]]
    )
    deploy_timer_event.start()
    update_timer_event.start()
    disable_timer_event.start()

    captured_events = []
    for event in dm.watch_objects(
        kind, api_version, namespace=SOME_OTHER_NAMESPACE, timeout=3
    ):
        captured_events.append(event)

    assert len(captured_events) == 4
    assert captured_events[0].type == KubeEventType.ADDED
    assert captured_events[0].resource.get("metadata").get("name") == "foobar"
    assert captured_events[1].type == KubeEventType.ADDED
    assert captured_events[1].resource.get("metadata").get("name") == "barfoo"
    assert captured_events[2].type == KubeEventType.MODIFIED
    assert captured_events[2].resource.get("metadata").get("name") == "barfoo"
    assert captured_events[3].type == KubeEventType.DELETED
    assert captured_events[3].resource.get("metadata").get("name") == "foobar"


@pytest.mark.timeout(5)
def test_watch_objects_namespaced():
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    name = "foobar"
    obj = make_obj(api_version, kind, name=name)
    dm.deploy([obj])

    update_timer_event = Timer(
        interval=0.5,
        function=dm.deploy,
        args=[[make_obj(api_version, kind, spec={"updated": 2})]],
    )
    other_update_timer_event = Timer(
        interval=0.5,
        function=dm.deploy,
        args=[[make_obj(api_version, kind, name="othername", spec={"updated": 2})]],
    )
    disable_timer_event = Timer(
        interval=1.5, function=dm.disable, args=[[make_obj(api_version, kind)]]
    )
    other_update_timer_event.start()
    update_timer_event.start()
    disable_timer_event.start()

    captured_events = []
    for event in dm.watch_objects(
        kind, api_version, namespace=SOME_OTHER_NAMESPACE, name=name, timeout=3
    ):
        captured_events.append(event)

    assert len(captured_events) == 3
    assert captured_events[0].type == KubeEventType.ADDED
    assert captured_events[0].resource.get("metadata").get("name") == name
    assert captured_events[1].type == KubeEventType.MODIFIED
    assert captured_events[1].resource.get("metadata").get("name") == name
    assert captured_events[2].type == KubeEventType.DELETED
    assert captured_events[2].resource.get("metadata").get("name") == name


def test_pre_deploy_resources():
    """Make sure that the _watches member gets initialized before deploy is
    called when a deploy manager is initialized with pre-existing resources
    """
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    DryRunDeployManager(resources=[obj])


def test_no_watch_set_status():
    """Make sure that calling set_status doesn't trigger a watch"""
    dm = DryRunDeployManager()
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)

    def die(*_, **__):
        raise RuntimeError()

    dm.register_watch(api_version, kind, die)
    dm.set_status(
        kind=obj["kind"],
        name=obj["metadata"]["name"],
        namespace=obj["metadata"]["namespace"],
        api_version=obj["apiVersion"],
        status={"foo": "bar"},
    )


def test_deploy_add_owner_reference():
    """Make sure that deploying a resource with an owner CR adds the owner
    reference given the resource and owner CR are in the same namespace
    """
    owner_cr = setup_cr(namespace=TEST_NAMESPACE)
    dm = DryRunDeployManager(owner_cr=owner_cr)
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind, namespace=TEST_NAMESPACE)
    success, changed = dm.deploy([obj])
    assert success
    assert changed
    success, content = dm.get_object_current_state(
        kind=kind,
        api_version=api_version,
        namespace=obj.metadata.namespace,
        name=obj.metadata.name,
    )
    assert success
    assert content["metadata"]["ownerReferences"] == [_make_owner_reference(owner_cr)]


def test_deploy_dont_add_owner_reference_if_manage_owner_references_disabled():
    """Make sure that deploying a resource with an owner CR does not add the
    owner reference when manage_owner_references is set to False
    """
    owner_cr = setup_cr(namespace=TEST_NAMESPACE)
    dm = DryRunDeployManager(owner_cr=owner_cr)
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind)
    success, changed = dm.deploy([obj], manage_owner_references=False)
    assert success
    assert changed
    success, content = dm.get_object_current_state(
        kind=kind,
        api_version=api_version,
        namespace=obj.metadata.namespace,
        name=obj.metadata.name,
    )
    assert success
    assert "ownerReferences" not in content["metadata"]


def test_deploy_dont_add_owner_reference_if_resource_and_owner_cr_not_in_same_namespace():
    """Make sure that deploying a resource with an owner CR does not add the owner
    reference if the resource and owner CR are in a different namespace
    """
    owner_cr = setup_cr(namespace=TEST_NAMESPACE)
    dm = DryRunDeployManager(owner_cr=owner_cr)
    api_version = "foo.bar/v1"
    kind = "Foo"
    obj = make_obj(api_version, kind, namespace=SOME_OTHER_NAMESPACE)
    success, changed = dm.deploy([obj])
    assert success
    assert changed
    success, content = dm.get_object_current_state(
        kind=kind,
        api_version=api_version,
        namespace=obj.metadata.namespace,
        name=obj.metadata.name,
    )
    assert success
    assert content["metadata"]["ownerReferences"] == []


def test_deploy_cluster_resource():
    """Make sure that deploying a resource without a namespace does not fail"""
    # Create deploy manager
    owner_cr = setup_cr()
    dm = DryRunDeployManager(owner_cr=owner_cr)
    api_version = "foo.bar/v1"
    kind = "Foo"

    # Create object and delete namespace
    obj = make_obj(api_version, kind)
    del obj["metadata"]["namespace"]

    # Deploy object
    success, changed = dm.deploy([obj], manage_owner_references=False)
    assert success
    assert changed

    success, content = dm.get_object_current_state(
        kind=kind,
        api_version=api_version,
        namespace=None,
        name=obj.metadata.name,
    )

    # Validate object was deployed
    assert success
    assert content["metadata"]["name"] == obj["metadata"]["name"]


@pytest.mark.parametrize(
    ["resources", "label_selector", "field_selector", "obj_count"],
    [
        # Label Selector Tests
        [[make_obj("foo/v1", "Foo", name="first")], "app==foobar", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app==wrong", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "app!=wrong", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app!=foobar", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "app=foobar", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app=wrong", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "app in foobar", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app in (foobar)", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app in (foobar,wrong)", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app in (wrong)", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "app in ()", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "app notin foobar", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "app notin (foobar)", None, 0],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            "app notin (foobar,wrong)",
            None,
            0,
        ],
        [[make_obj("foo/v1", "Foo", name="first")], "app notin (wrong)", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app notin ()", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "app", None, 1],
        [[make_obj("foo/v1", "Foo", name="first")], "wrong", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "!app", None, 0],
        [[make_obj("foo/v1", "Foo", name="first")], "!wrong", None, 1],
        # Label Selector MultiResult Tests
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            "app==foobar",
            None,
            2,
        ],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            "app in foobar",
            None,
            2,
        ],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            "app not in foobar",
            None,
            0,
        ],
        # Label Selector with Multiple Label Selectors
        [[make_obj("foo/v1", "Foo", name="first")], "app,app==foobar", None, 1],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            "app in (foobar),run=frontend",
            None,
            1,
        ],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            "app in (foobar),run!=backend,owner=prod",
            None,
            0,
        ],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            "app,app in (foobar,barfoo),run,run notin backend",
            None,
            1,
        ],
        # Field Selector Tests
        # no need to test so many permutations as they're covered in label selectors Tests
        [[make_obj("foo/v1", "Foo", name="first")], None, "metadata.name==first", 1],
        [[make_obj("foo/v1", "Foo", name="first")], None, "metadata.name==second", 0],
        [[make_obj("foo/v1", "Foo", name="first")], None, "spec.a!=1", 0],
        [[make_obj("foo/v1", "Foo", name="first")], None, "spec.a=1", 1],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            None,
            "metadata.labels.app in (foobar)",
            1,
        ],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            None,
            "metadata.labels.app in (wrong)",
            0,
        ],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            None,
            "metadata.labels.app notin (wrong)",
            1,
        ],
        [[make_obj("foo/v1", "Foo", name="first")], None, "!spec.b", 1],
        [[make_obj("foo/v1", "Foo", name="first")], None, "spec.a", 1],
        # Field Selector Multi Tests
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            None,
            "metadata.name = first",
            1,
        ],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            None,
            "metadata.name in (first,second)",
            2,
        ],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            None,
            "metadata.namespace notin (default)",
            2,
        ],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            None,
            "spec.a",
            2,
        ],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            None,
            "spec.a=4",
            0,
        ],
        # Label And Field Selector Combined
        [[make_obj("foo/v1", "Foo", name="first")], "app=foobar", "spec.a=1", 1],
        [
            [make_obj("foo/v1", "Foo", name="first")],
            "app in (foobar,barfoo)",
            "spec.a=2",
            0,
        ],
        [[make_obj("foo/v1", "Foo", name="first")], "app in (wrong)", "spec.a=1", 0],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            "app=foobar",
            "metadata.name==first",
            1,
        ],
        # Blank Field and Label Selectors match all
        [[make_obj("foo/v1", "Foo", name="first")], None, None, 1],
        [
            [
                make_obj("foo/v1", "Foo", name="first"),
                make_obj("foo/v1", "Foo", name="second"),
            ],
            None,
            None,
            2,
        ],
    ],
)
def test_filter_objects_current_state(
    resources, label_selector, field_selector, obj_count
):
    """Make sure that all types of field selectors and label selectors work as expected"""
    # Create deploy manager
    owner_cr = setup_cr()
    dm = DryRunDeployManager(owner_cr=owner_cr)
    api_version = resources[0].get("apiVersion")
    namespace = resources[0].get("metadata").get("namespace")
    kind = resources[0].get("kind")

    # Deploy object
    command_success, changed = dm.deploy(resources, manage_owner_references=False)
    assert command_success
    assert changed

    # Filter Object
    command_success, content = dm.filter_objects_current_state(
        kind,
        api_version=api_version,
        namespace=namespace,
        label_selector=label_selector,
        field_selector=field_selector,
    )

    assert command_success
    assert len(content) == obj_count


def test_filter_object_state_incorrect_api_version():
    """Make sure that deploying a resource without a namespace does not fail"""
    # Create deploy manager
    owner_cr = setup_cr()
    dm = DryRunDeployManager(owner_cr=owner_cr)
    api_version = "wrong"
    namespace = "default"
    kind = "Foo"

    resources = [make_obj("right", "Foo", namespace=namespace, name="first")]

    # Deploy object
    command_success, changed = dm.deploy(resources, manage_owner_references=False)
    assert command_success
    assert changed

    # Filter Object
    command_success, content = dm.filter_objects_current_state(
        kind,
        api_version=api_version,
        namespace=namespace,
        label_selector=None,
        field_selector=None,
    )

    assert command_success
    assert len(content) == 0



"""
This module implements Patch Strategic Merge following the semantics in:

* kustomize: https://kubectl.docs.kubernetes.io/references/kustomize/glossary/#patchstrategicmerge
* kubernetes: https://github.com/kubernetes/community/blob/master/contributors/devel/sig-api-machinery/strategic-merge-patch.md
"""  # pylint: disable=line-too-long


# Standard
from collections import OrderedDict
from typing import Dict
import copy

# Third Party
from openshift.dynamic.apply import STRATEGIC_MERGE_PATCH_KEYS

# First Party
import alog

log = alog.use_channel("PATCH")

## Public ######################################################################


def patch_strategic_merge(
    resource_definition: dict,
    patch: dict,
    merge_patch_keys: Dict[str, str] = None,
) -> dict:
    """Apply a Strategic Merge Patch based on JSON Merge Patch (rfc 7386)

    Args:
        resource_definition:  dict
            The dict representation of the kubernetes resource
        patch:  dict
            The formatted patch to apply
        merge_patch_keys:  Dict[str, str]
            The mapping from paths to merge keys used to perform merge semantics
            for list elements

    Returns:
        patched_resource_definition:  dict
            The patched version of the resource_definition
    """
    if merge_patch_keys is None:
        merge_patch_keys = STRATEGIC_MERGE_PATCH_KEYS
    return _strategic_merge(
        current=copy.deepcopy(resource_definition),
        desired=copy.deepcopy(patch),
        position=resource_definition.get("kind"),
        merge_patch_keys=merge_patch_keys,
    )


## Implementation ##############################################################


_DIRECTIVE_KEY = "$patch"
_DIRECTIVE_REPLACE = "replace"
_DIRECTIVE_MERGE = "merge"
_DIRECTIVE_DELETE = "delete"
_DIRECTIVE_DELETE_FROM_PRIMITIVE_LIST = "$deleteFromPrimitiveList/"


def _strategic_merge(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    current: dict,
    desired: dict,
    position: str,
    merge_patch_keys: Dict[str, str],
) -> dict:
    """Recursive implementation of Patch Strategic Merge."""

    # If they are dicts, recurse on keys
    if isinstance(desired, dict) and isinstance(current, dict):
        log.debug4("Performing dict merge at [%s]", position)
        log.debug4(current)
        log.debug4(desired)
        for key, val in desired.items():
            # Support deletion
            if val is None:
                current.pop(key, None)

            # Check for the special directive to remove from a primitive list
            elif key.startswith(_DIRECTIVE_DELETE_FROM_PRIMITIVE_LIST):
                target_key = key.split("/", 1)[-1]
                if target_key not in current:
                    raise ValueError(
                        f"Cannot delete from unknown primitive list [{target_key}]"
                    )
                target_val = current[target_key]
                if not isinstance(val, list):
                    raise ValueError(
                        "Bad primitive list delete directive. Patch must be a list."
                    )
                if not isinstance(target_val, list):
                    raise ValueError(
                        "Bad primitive list delete directive. Target must be a list."
                    )
                for element in val:
                    try:
                        target_val.remove(element)
                    except ValueError as err:
                        raise ValueError(
                            "Bad primitive list delete directive. Element not found."
                        ) from err

            # Add new keys
            elif key not in current:
                current[key] = val

            # Recurse
            else:
                next_position = ".".join([position, key])
                log.debug4("Recursing to [%s]", next_position)
                current[key] = _strategic_merge(
                    current[key],
                    val,
                    next_position,
                    merge_patch_keys,
                )

        return current

    # If they are lists, apply the strategic merge
    if isinstance(desired, list) and isinstance(current, list):
        merge_key = merge_patch_keys.get(position)
        log.debug4("Performing list merge at [%s]. Merge key: %s", position, merge_key)

        # If no merge key given for this path, do an overwrite merge
        if not merge_key:
            return desired

        # Otherwise, align elements by their merge keys and recursively merge
        # each one

        # Make sure these are lists of dicts
        if not all(isinstance(itm, dict) and merge_key in itm for itm in current):
            raise ValueError(
                f"Current at [{position}] contains elements without [{merge_key}]"
            )
        if not all(isinstance(itm, dict) and merge_key in itm for itm in desired):
            raise ValueError(
                f"Desired at [{position}] contains elements without [{merge_key}]"
            )

        # Create dicts based on the merge key
        current_dict = OrderedDict([(itm[merge_key], itm) for itm in current])
        desired_dict = OrderedDict([(itm[merge_key], itm) for itm in desired])

        # Perform the merge on each item
        for item_key, item in desired_dict.items():
            # Get the directive (default to merge)
            directive = item.pop(_DIRECTIVE_KEY, _DIRECTIVE_MERGE)
            log.debug4("Element [%s] directive: %s", item_key, directive)

            # If "delete" remove from the current
            if directive == _DIRECTIVE_DELETE:
                log.debug4("Doing delete")
                if item_key not in current_dict:
                    raise ValueError(
                        f"Invalid [{_DIRECTIVE_DELETE}] on missing element [{item_key}]"
                    )
                del current_dict[item_key]

            # If "replace" just replace in the current
            elif directive == _DIRECTIVE_REPLACE or item_key not in current_dict:
                log.debug4("Doing replace")
                current_dict[item_key] = item

            # If "merge" recurse
            elif directive == _DIRECTIVE_MERGE:
                # NOTE: list nesting is not represented in the position for
                #   merge keys
                current_dict[item_key] = _strategic_merge(
                    current_dict[item_key],
                    item,
                    position,
                    merge_patch_keys,
                )

            # Otherwise, it's an error
            else:
                raise ValueError(f"Invalid directive: [{directive}]")

        # Return the ordered list of updated elements
        return list(current_dict.values())

    # If not one of the special types, overwrite
    log.debug4("Performing overwrite")
    return desired



"""
This module holds shared semantics for patching resources using temporary_patch
"""

# Standard
from typing import List
import copy

# Third Party
from jsonpatch import JsonPatch

# First Party
import alog

# Local
from .patch_strategic_merge import patch_strategic_merge

log = alog.use_channel("PATCH")

## Public Interface ############################################################

STRATEGIC_MERGE_PATCH = "patchStrategicMerge"
JSON_PATCH_6902 = "patchJson6902"


def apply_patches(
    internal_name: str,
    resource_definition: dict,
    temporary_patches: List[dict],
):
    """Apply all temporary patches to the given resource from the given list.
    The patches are applied in-place.

    Args:
        internal_name:  str
            The name given to the internal node of the object. This is used to
            identify which patches apply to this object.
        resource_definition:  dict
            The dict representation of the object to patch
        temporary_patches:  List[dict]
            The list of temporary patches that apply to this rollout

    Returns:
        patched_definition:  dict
            The dict representation of the object with patches applied
    """
    log.debug2(
        "Looking for patches for %s/%s (%s)",
        resource_definition.get("kind"),
        resource_definition.get("metadata", {}).get("name"),
        internal_name,
    )
    resource_definition = copy.deepcopy(resource_definition)
    for patch_content in temporary_patches:
        log.debug4("Checking patch: << %s >>", patch_content)

        # Look to see if this patch contains a match for the internal name
        internal_name_parts = internal_name.split(".")
        internal_name_parts.reverse()
        patch = patch_content.spec.patch
        log.debug4("Full patch section: %s", patch)
        while internal_name_parts and isinstance(patch, dict):
            patch_level = internal_name_parts.pop()
            log.debug4("Getting patch level [%s]", patch_level)
            patch = patch.get(patch_level, {})
            log.debug4("Patch level: %s", patch)
        log.debug4("Checking patch: %s", patch)

        # If the patch matches, apply the right merge
        if patch and not internal_name_parts:
            log.debug3("Found matching patch: %s", patch_content.metadata.name)

            # Dispatch the right patch type
            if patch_content.spec.patchType == STRATEGIC_MERGE_PATCH:
                resource_definition = _apply_patch_strategic_merge(
                    resource_definition, patch
                )
            elif patch_content.spec.patchType == JSON_PATCH_6902:
                resource_definition = _apply_json_patch(resource_definition, patch)
            else:
                raise ValueError(
                    f"Unsupported patch type [{patch_content.spec.patchType}]"
                )
    return resource_definition


## JSON Patch 6902 #############################################################


def _apply_json_patch(
    resource_definition: dict,
    patch: dict,
) -> dict:
    """Apply a Json Patch based on JSON Patch (rfc 6902)"""

    if not isinstance(patch, list):
        raise ValueError("Invalid JSON 6902 patch. Must be a list of operations.")
    return JsonPatch(patch).apply(resource_definition)


## Strategic Merge Patch #######################################################


def _apply_patch_strategic_merge(
    resource_definition: dict,
    patch: dict,
) -> dict:
    """Apply a Strategic Merge Patch based on JSON Merge Patch (rfc 7386)"""
    return patch_strategic_merge(resource_definition, patch)



"""
Helper module to implement a once-per-open log rotating handler
"""

# Standard
from logging.handlers import RotatingFileHandler
import os


class AutoRotatingFileHandler(RotatingFileHandler):
    """This subclass of the RotatingFileHandler will perform a rotation once on
    construction. The result is that the previous log file will be backed up
    before opening the new file, but the file will never rotate during a single
    reconcile loop.
    """

    def __init__(self, filename, backupCount=10):
        """Construct with only the filename and backupCount args of the base
        class.

        Args:
            filename:  str
                The name of the primary log file to manage
            backupCount:  int
                The number of backed up copies of the log file to keep
        """
        file_already_there = os.path.exists(filename)
        super().__init__(filename, backupCount=backupCount, maxBytes=0)
        if file_already_there:
            self.doRollover()



"""
The ReconcileThread is the heart of the PythonWatchManager and controls reconciling
resources and managing any subprocesses
"""
# Standard
from datetime import datetime
from logging.handlers import QueueListener
from multiprocessing.connection import Connection
from time import sleep
from typing import Dict, List, Optional
import multiprocessing
import multiprocessing.forkserver
import os
import queue
import threading
import time

# First Party
import alog

# Local
from .... import config
from ....deploy_manager import DeployManagerBase, KubeEventType
from ....exceptions import ConfigError
from ....managed_object import ManagedObject
from ....reconcile import ReconciliationResult
from ..filters import FilterManager
from ..leader_election import LeadershipManagerBase
from ..reconcile_process_entrypoint import create_and_start_entrypoint
from ..utils import (
    JOIN_PROCESS_TIMEOUT,
    SHUTDOWN_RECONCILE_POLL_TIME,
    ReconcileProcess,
    ReconcileRequest,
    ReconcileRequestType,
    Singleton,
    TimerEvent,
    WatchRequest,
    get_logging_handlers,
    parse_time_delta,
)
from .base import ThreadBase
from .timer import TimerThread
from .watch import create_resource_watch

log = alog.use_channel("RCLTHRD")


class ReconcileThread(
    ThreadBase, metaclass=Singleton
):  # pylint: disable=too-many-instance-attributes
    """This class is the core reconciliation class that handles starting subprocesses,
    tracking their status, and handles any results. This thread also kicks of requeue
    requests and requests dependent resource watches"""

    def __init__(
        self,
        deploy_manager: DeployManagerBase = None,
        leadership_manager: LeadershipManagerBase = None,
    ):
        """Initialize the required queues, helper threads, and reconcile tracking. Also
        gather any onetime configuration options

        Args:
            deploy_manager: DeployManagerBase = None
                The deploy manager used throughout the thread
            leadership_manager: LeadershipManagerBase = None
                The leadership_manager for tracking elections
        """
        super().__init__(
            name="reconcile_thread",
            deploy_manager=deploy_manager,
            leadership_manager=leadership_manager,
        )

        # Configure the multiprocessing process spawning context
        context = config.python_watch_manager.process_context
        if context not in multiprocessing.get_all_start_methods():
            raise ConfigError(f"Invalid process_context: '{context}'")

        if context == "fork":
            log.warning(
                "The fork multiprocessing context is known to cause deadlocks in certain"
                " environments due to OpenSSL. Consider using spawn for more reliable"
                " reconciling"
            )

        self.spawn_ctx = multiprocessing.get_context(context)

        # Setup required queues
        self.request_queue = self.spawn_ctx.Queue()
        self.logging_queue = self.spawn_ctx.Queue()

        # Setup helper threads
        self.timer_thread: TimerThread = TimerThread()
        self.log_listener_thread: QueueListener = QueueListener(
            self.logging_queue, *get_logging_handlers(), respect_handler_level=False
        )

        # Setup reconcile, request, and event mappings
        self.running_reconciles: Dict[str, ReconcileProcess] = {}
        self.pending_reconciles: Dict[str, ReconcileRequest] = {}
        self.event_map: Dict[str, TimerEvent] = {}

        # Setup control variables
        self.process_overload = threading.Event()

        # Configure the max number of concurrent reconciles via either config
        # or number of cpus
        if config.python_watch_manager.max_concurrent_reconciles:
            self.max_concurrent_reconciles = (
                config.python_watch_manager.max_concurrent_reconciles
            )
        else:
            self.max_concurrent_reconciles = os.cpu_count()

    def run(self):
        """The reconcile threads control flow is to first wait for
        either a new reconcile request or for a process to end. If its a reconcile request
        the thread checks if one is already running for the resource and if not starts a
        new one. If a reconcile is already running or the thread couldn't start a new one
        the request gets added to the pending reconciles. There can only be one pending
        reconcile per resource. If the reconcile thread received a process end event it
        checks the exit code and handles the result queue.
        """
        while True:  # pylint: disable=too-many-nested-blocks
            if not self.check_preconditions():
                return

            # Wait for a change with the reconcile processes or reconcile queue.
            # Use _reader and Process.sentinel objects, so we can utilize select.select
            listen_objs = [
                *self.running_reconciles.values(),
                self.request_queue._reader,  # pylint: disable=protected-access
            ]
            log.debug3("Waiting on %s", listen_objs)
            ready_objs = multiprocessing.connection.wait(listen_objs)

            # Check preconditions both before and after waiting
            if not self.check_preconditions():
                return

            # For each object that triggered the connection
            for obj in ready_objs:
                log.debug3("Processing object %s with type %s", obj, type(obj))

                # Handle reconcile process end events
                if isinstance(obj, ReconcileProcess):
                    if self._handle_process_end(obj):
                        # If process overload is set than we need to check all resources for
                        # pending reconciles otherwise just check if the completed resource
                        # has a pending request.
                        if self.process_overload.is_set():
                            for uid in list(self.pending_reconciles.keys()):
                                if not self._handle_pending_reconcile(uid):
                                    break
                        else:
                            self._handle_pending_reconcile(obj.uid())

                # Handle all of the events in the queue
                elif isinstance(obj, Connection):
                    self._handle_request_queue()

    ## Class Interface ###################################################

    def start_thread(self):
        """Override start_thread to start helper threads"""
        self.timer_thread.start_thread()
        self.log_listener_thread.start()
        super().start_thread()

    def stop_thread(self):
        """Override stop_thread to ensure reconciles finish correctly"""
        super().stop_thread()

        if not self.is_alive() and not self.running_reconciles:
            log.debug("Reconcile Thread already stopped")
            return

        # Reawaken reconcile thread to stop
        log.debug("Pushing stop reconcile request")
        self.push_request(ReconcileRequest(None, ReconcileRequestType.STOPPED, {}))

        log.debug("Waiting for reconcile thread to finish")
        while self.is_alive():
            time.sleep(0.001)

        # Wait until all reconciles have completed
        log.info("Waiting for Running Reconciles to end")
        while self.running_reconciles:
            log.debug2("Waiting for reconciles %s to end", self.running_reconciles)
            for reconcile_process in list(self.running_reconciles.values()):
                # attempt to join process before trying the next one
                reconcile_process.process.join(JOIN_PROCESS_TIMEOUT)
                if reconcile_process.process.exitcode is not None:
                    log.debug(
                        "Joined reconciles process %s with exitcode: %s for request object %s",
                        reconcile_process.process.pid,
                        reconcile_process.process.exitcode,
                        reconcile_process.request,
                    )
                    self.running_reconciles.pop(reconcile_process.uid())
                    reconcile_process.process.close()

            # Pause for slight delay between checking processes
            sleep(SHUTDOWN_RECONCILE_POLL_TIME)

        # Close the logging queue to indicate no more logging events
        self.logging_queue.close()
        # Skip stopping the listener thread as it can hang on the join, this isn't
        # too bad as the listener thread is daemon anyways
        # self.log_listener_thread.stop()

    ## Public Interface ###################################################

    def push_request(self, request: ReconcileRequest):
        """Push request to reconcile queue

        Args:
            request: ReconcileRequest
                the ReconcileRequest to add to the queue
        """
        log.info(
            "Pushing request '%s' to reconcile queue",
            request,
            extra={"resource": request.resource},
        )
        self.request_queue.put(request)

    ## Event Handlers ###################################################

    def _handle_request_queue(self):
        """The function attempts to start a reconcile for every reconcile requests in the queue.
        If it can't start a reconcile or one is already running then it pushes it to the pending
        queue"""

        # Get all events from the queue
        pending_requests = self._get_all_requests()

        # Start a reconcile for each pending request
        for request in pending_requests:
            if request.type == ReconcileRequestType.STOPPED:
                break

            log.debug3("Got request %s from queue", request)

            # If reconcile is not running then start the process. Otherwise
            # or if starting failed push to the pending reconcile queue
            if request.resource.uid not in self.running_reconciles:
                if not self._start_reconcile_for_request(request):
                    self._push_to_pending_reconcile(request)
            else:
                self._push_to_pending_reconcile(request)

    def _handle_process_end(self, reconcile_process: ReconcileProcess) -> str:
        """Handle a process end event. The function joins the finished process,
        manages any events in the pipe, and creates a requeue/periodic event if
        one is needed.

        Args:
            reconcile_process: ReconcileProcess
                The process that ended

        Returns:
            uid: str
                The uid of the resource that ended
        """
        # Parse process variables
        uid = reconcile_process.uid()
        reconcile_request = reconcile_process.request
        process = reconcile_process.process
        pipe = reconcile_process.pipe

        # Attempt to join the process
        log.debug(
            "Joining process for request %s",
            reconcile_request,
            extra={"resource": reconcile_request.resource},
        )
        process.join(JOIN_PROCESS_TIMEOUT)
        exit_code = process.exitcode

        # If its still then exit and process   will be cleaned up on next iteration
        if exit_code is None:
            log.debug("Process is still alive after join. Skipping cleanup")
            return None

        if exit_code != 0:
            log.warning(
                "Reconcile did not complete successfully: %s",
                reconcile_request,
                extra={"resource": reconcile_request.resource},
            )

        # Remove reconcile from map and release resource lock
        self.running_reconciles.pop(uid)
        self.leadership_manager.release_resource(reconcile_request.resource)

        # Handle any events passed via the process pipe including the reconcile result and
        # close the pipe once done
        reconcile_result = self._handle_process_pipe(pipe)
        process.close()

        # Print reconciliation result
        log.info(
            "Reconcile completed with result %s",
            reconcile_result if reconcile_result else exit_code,
            extra={"resource": reconcile_request.resource},
        )

        # Cancel any existing requeue events
        if uid in self.event_map:
            log.debug2("Marking event as stale: %s", self.event_map[uid])
            self.event_map[uid].cancel()

        # Create a new timer event if one is needed
        event = self._create_timer_event_for_request(
            reconcile_request, reconcile_result
        )
        if event:
            self.event_map[uid] = event

        return uid

    def _handle_process_pipe(self, pipe: Connection) -> ReconciliationResult:
        """Handle any objects in a connection pipe and return the reconciliation result

        Args:
            pipe: Connection
                the pipe to read results from

        Returns:
            reconcile_result: ReconciliationResult
                The result gathered from the pipe
        """
        reconcile_result = None
        while pipe.poll():
            # EOFError is raised when the pipe is closed which is expected after the reconcile
            # process has been joined
            try:
                pipe_obj = pipe.recv()
            except EOFError:
                break

            log.debug3("Received obj %s from process pipe", pipe_obj)

            # Handle any watch requests received
            if isinstance(pipe_obj, WatchRequest):
                self._handle_watch_request(pipe_obj)

            # We only expect one reconciliation result per process
            elif isinstance(pipe_obj, ReconciliationResult):
                reconcile_result = pipe_obj

        # Close the reconcile pipe and release the rest of the process resources
        pipe.close()

        return reconcile_result

    def _handle_watch_request(self, request: WatchRequest):
        """Create a resource watch for a given watch request. This function also
        handles converting controller_info into a valid controller_type

        Args:
            request: WatchRequest
                The requested WatchRequest to be created
        """
        # Parse the controller info into a type
        if request.controller_info and not request.controller_type:
            request.controller_type = request.controller_info.to_class()

        # Parse any filter infos into types
        if request.filters_info:
            request.filters = FilterManager.from_info(request.filters_info)

        create_resource_watch(
            request,
            self,
            self.deploy_manager,
            self.leadership_manager,
        )

    def _create_timer_event_for_request(
        self, request: ReconcileRequest, result: ReconciliationResult = None
    ) -> Optional[TimerEvent]:
        """Enqueue either a requeue or periodic reconcile request for a given
        result.

        Args:
            request: ReconcileRequest
                The original reconcile request that triggered this process
            result: ReconciliationResult = None
                The result of the reconcile

        Returns:
            timer_event: Optional[TimerEvent]
                The timer event if one was created
        """

        # Short circuit if event is not needed, if resource was deleted,
        # or if theres already a pending reconcile
        if (
            not result or not result.requeue
        ) and not config.python_watch_manager.reconcile_period:
            return None

        if result and not result.requeue and request.type == KubeEventType.DELETED:
            return None

        if request.resource.uid in self.pending_reconciles:
            return None

        # Create requeue_time and type based on result/config
        request_type = None
        requeue_time = None
        if result and result.requeue:
            requeue_time = datetime.now() + result.requeue_params.requeue_after
            request_type = ReconcileRequestType.REQUEUED
        elif config.python_watch_manager.reconcile_period:
            requeue_time = datetime.now() + parse_time_delta(
                config.python_watch_manager.reconcile_period
            )
            request_type = ReconcileRequestType.PERIODIC

        future_request = ReconcileRequest(
            request.controller_type, request_type, request.resource
        )
        log.debug3("Pushing requeue request to timer: %s", future_request)

        return self.timer_thread.put_event(
            requeue_time, self._push_updated_request, future_request
        )

    def _push_updated_request(self, request: ReconcileRequest):
        """_push_updated_request is a helper function to ensure that when submitting a requeue
        event we always fetch the latest resource.

        Args:
            request (ReconcileRequest): The request to push
        """

        success, current_resource = self.deploy_manager.get_object_current_state(
            kind=request.resource.kind,
            api_version=request.resource.api_version,
            name=request.resource.name,
            namespace=request.resource.namespace,
        )
        if not success:
            log.debug2("Unable to locate resource for requeue request: %s", request)
            return

        request.resource = ManagedObject(current_resource)
        self.push_request(request)

    ## Pending Event Helpers ###################################################

    def _handle_pending_reconcile(self, uid: str) -> bool:
        """Start reconcile for pending request if there is one

        Args:
             uid:str
                The uid of the resource being reconciled

        Returns:
            successful_start:bool
                If there was a pending reconcile that got started"""
        # Check if resource has pending request
        if uid in self.running_reconciles or uid not in self.pending_reconciles:
            return False

        # Start reconcile for request
        request = self.pending_reconciles[uid]
        log.debug4("Got request %s from pending reconciles", request)
        if self._start_reconcile_for_request(request):
            self.pending_reconciles.pop(uid)
            return True
        return False

    def _push_to_pending_reconcile(self, request: ReconcileRequest):
        """Push a request to the pending queue if it's newer than the current event

        Args:
            request:  ReconcileRequest
                The request to possibly add to the pending_reconciles
        """
        uid = request.uid()
        # Only update queue if request is newer
        if uid in self.pending_reconciles:
            if request.timestamp > self.pending_reconciles[uid].timestamp:
                log.debug3("Updating reconcile queue with event %s", request)
                self.pending_reconciles[uid] = request
            else:
                log.debug4("Event in queue is newer than event %s", request)
        else:
            log.debug3("Adding event %s to reconcile queue", request)
            self.pending_reconciles[uid] = request

    ## Process functions ##################################################

    def _start_reconcile_for_request(self, request: ReconcileRequest) -> bool:
        """Start a reconciliation process for a given request

        Args:
            request: ReconcileRequest
                The request to attempt to start

        Returns:
            successfully_started: bool
                If a process could be started
        """
        # If thread is supposed to shutdown then don't start process
        if self.should_stop():
            return False

        # Check if there are too many reconciles running
        if len(self.running_reconciles.keys()) >= self.max_concurrent_reconciles:
            log.warning("Unable to start reconcile, max concurrent jobs reached")
            self.process_overload.set()
            return False

        # Attempt to acquire lock on resource. If failed skip starting
        if not self.leadership_manager.acquire_resource(request.resource):
            log.debug("Unable to obtain leadership lock for %s", request)
            return False

        self.process_overload.clear()
        log.info(
            "Starting reconcile for request %s",
            request,
            extra={"resource": request.resource},
        )

        # Create the send and return pipe
        recv_pipe, send_pipe = self.spawn_ctx.Pipe()

        process = self._start_reconcile_process(request, send_pipe)

        # Generate the reconcile process and update map
        reconcile_process = ReconcileProcess(
            process=process, request=request, pipe=recv_pipe
        )

        self.running_reconciles[request.uid()] = reconcile_process
        return True

    def _start_reconcile_process(
        self, request: ReconcileRequest, pipe: Connection
    ) -> multiprocessing.Process:
        """Helper function to generate and start the reconcile process. This
        was largely created to ease the testing and mocking process

         Args:
             request: ReconcileRequest
                The request to start the process with
             pipe: Connection
                The result pipe for this reconcile

        Returns:
            process: multiprocessing.Process
                The started process
        """

        # Create and start the reconcile process
        process = self.spawn_ctx.Process(
            target=create_and_start_entrypoint,
            args=[self.logging_queue, request, pipe],
        )
        process.start()
        log.debug3("Started child process with pid: %s", process.pid)
        return process

    ## Queue Functions ##################################################

    def _get_all_requests(
        self, timeout: Optional[int] = None
    ) -> List[ReconcileRequest]:
        """Get all of the requests from the reconcile queue

        Args:
            timeout:Optional[int]=None
                The timeout to wait for an event. If None it returns immediately

        Returns:
            requests: List[ReconcileRequest]
                The list of requests gathered from the queue
        """
        request_list = []
        while not self.request_queue.empty():
            try:
                request = self.request_queue.get(block=False, timeout=timeout)

                # If there is a stop request then immediately return it
                if request.type == ReconcileRequestType.STOPPED:
                    return [request]
                request_list.append(request)
            except queue.Empty:
                break
        return request_list



"""
This module contains a collection of classes for executing functions along a DAG
"""

# Standard
from concurrent.futures import Executor, Future, ThreadPoolExecutor
from typing import Callable, List, Optional
import time

# First Party
import alog

# Local
from .completion_state import CompletionState
from .graph import Graph
from .node import Node

log = alog.use_channel("Runner")

## Executer ##############################################################


class Runner:  # pylint: disable=too-many-instance-attributes
    """This is a very simple "keep running until done" Runner executor which uses
    a ThreadPoolExecutor to allow non-blocking calls to execute in parallel.
    """

    @property
    def graph(self) -> str:  # pylint: disable=missing-function-docstring
        return self._graph

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str = "",
        threads: Optional[int] = None,
        graph: Optional[Graph] = None,
        default_function: Optional[Callable[["Node"], bool]] = None,
        poll_time: float = 0.05,
        verify_upstream: bool = True,
    ):
        """Construct a Runner which will manage the execution of a single Graph

        Args:
            name:  str
                String name that can be used for logging to differentiate deploy
                and verify Graph executions
            threads:  Optional[int]
                Number of threads to use. If not given, the default behavior of
                ThreadPoolExecutor is to use the number of available cores.
            graph:  Optional[Graph]
                Existing graph to use, if not supplied an empty graph is created
            default_function: Optional[Callable[["Node"],None]]=None
                Function that will be called on node run if specific function is not provided
            poll_time:  float
                How often to check runner status
            verify_upstream: bool
                Conditional to control whether to check a nodes upstream via its edge function
        """
        self.name = name
        # If threads are disabled, use the NonThreadPoolExecutor
        if threads == 0:
            log.debug("Running without threading")
            pool_type = NonThreadPoolExecutor
        else:
            log.debug("Running with %s threads", threads)
            pool_type = ThreadPoolExecutor
        self._pool = pool_type(max_workers=threads)
        self._graph = graph or Graph()
        self._default_node_func = default_function or (lambda _: None)

        self._failed = False
        self._exception = None
        self._verify_upstream = verify_upstream
        self._poll_time = poll_time
        self._started_nodes = []
        self._disabled_nodes = []

        # Nodes can terminate in one of three states:
        #   1. Completed and verified
        #   2. Completed, but not verified
        #   3. Failed
        self._verified_nodes = []
        self._unverified_nodes = []
        self._failed_nodes = []

    ## Public ##

    def disable_node(
        self,
        node: "Node",
    ):
        """Function to disable a node in the graph. This will skip the node in runner without
        changing the graph"""
        graph_node = self.graph.get_node(node.get_name())
        if graph_node:
            self._disabled_nodes.append(graph_node)

    def enable_node(
        self,
        node: "Node",
    ):
        """Function to reenable a node after it was disabled by Runner.disable_node"""
        graph_node = self.graph.get_node(node.get_name())
        if graph_node in self._disabled_nodes:
            self._disabled_nodes.remove(graph_node)

    def completion_state(self):
        """Get the state of which nodes completed and which failed

        Returns:
            completion_state:  CompletionState
                The state holding the full view of the termination state of each
                node
        """
        return CompletionState(
            verified_nodes=self._verified_nodes,
            unverified_nodes=self._unverified_nodes,
            failed_nodes=self._failed_nodes,
            unstarted_nodes=[
                node
                for node in self.graph.get_all_nodes()
                if node not in self._get_completed_nodes()
            ],
            exception=self._exception,
        )

    def run(self):
        """Run the Runner! This will continue until the graph has run to completion
        or halted due to an error.
        """
        node_list = self._get_runnable_nodes()
        log.debug3(
            "Started Nodes: %s, All Nodes: %s",
            self._started_nodes,
            list(node_list),
        )

        # The "node start" loop should terminate if:
        # 1. All nodes have started
        # 2. All started nodes have completed in one form or another and there
        #   are no newly ready nodes
        while len(self._started_nodes) < len(node_list):
            # Get the set of nodes that has completed already
            #
            # NOTE: It's _critically_ important that this be done before getting
            #   the ready nodes. The operation of getting ready nodes can
            #   delegate to user-defined verification functions which may be
            #   very slow and IO bound. With slow verification functions, a node
            #   running in a thread may complete and mark itself verified after
            #   a downstream dependency has checked its completion status, but
            #   before the full set of _get_ready_nodes() checks has passed. If
            #   this happens and _get_completed_nodes() is called afterwards,
            #   the short-circuit logic below will think that all started nodes
            #   have completed and there are no ready nodes, thus terminating
            #   the Runner prematurely.
            completed_nodes = self._get_completed_nodes()

            # Get the currently ready nodes
            ready_nodes = self._get_ready_nodes()

            # If there are no ready nodes and all started nodes have completed
            # in one way or another, we're in an early termination case
            log.debug4("Ready Nodes: %s", ready_nodes)
            log.debug4("Completed Nodes: %s", completed_nodes)
            if not ready_nodes and set(self._started_nodes) == set(completed_nodes):
                log.debug2(
                    "[%s] Graph exhausted all available nodes. Terminating early.",
                    self.name,
                )
                break

            # If there are new ready nodes, start them
            if ready_nodes:
                log.debug2(
                    "Ready nodes: %s. Remaining nodes: %s",
                    ready_nodes,
                    [
                        node
                        for node in node_list
                        if node not in ready_nodes and node not in completed_nodes
                    ],
                )
            for ready_node in ready_nodes:
                self._started_nodes.append(ready_node)
                self._pool.submit(self._run_node, ready_node)
            time.sleep(self._poll_time)

        # Log out the state of the graph once we've terminated, but before we've
        # waited for all nodes to terminate
        log.debug2("[NODES] Started: %s", sorted(self._started_nodes))
        log.debug2("[NODES] Verified: %s", sorted(self._verified_nodes))
        log.debug2("[NODES] Unverified: %s", sorted(self._unverified_nodes))
        log.debug2("[NODES] Failed: %s", sorted(self._failed_nodes))
        log.debug2("[NODES] All: %s", sorted(list(node_list)))

        # Wait until all started nodes have finished one way or the other
        while len(self._get_completed_nodes()) != len(self._started_nodes):
            time.sleep(self._poll_time)

        # Make sure any in-flight nodes complete before terminating
        log.debug2("Waiting for in-flight nodes to complete")
        self._pool.shutdown()
        log.debug2("All nodes complete")
        log.debug2(self.completion_state())

    ## Implementation Details ##

    def _run_node(self, node: "Node"):
        node_name = node.get_name()
        log.debug2("Starting node: %s", node_name)

        try:
            # Call node function or default
            node_func = node.get_data()
            if callable(node_func):
                node_func()
            else:
                self._default_node_func(node)

        except DagHaltError as err:
            log.debug("[%s] DagHaltError caught. Stopping Execution", self.name)
            self._failed = err.failure
            self._exception = err.exception
            if err.failure:
                self._failed_nodes.append(node)
            else:
                self._unverified_nodes.append(node)
        except Exception as err:  # pylint: disable=broad-except
            log.warning(
                "Unexpected exception caught in Runner node: %s", err, exc_info=True
            )
            self._failed = True
            self._failed_nodes.append(node)
        else:
            log.debug2("Node complete: %s", node_name)
            self._verified_nodes.append(node)

    def _dependency_satisfied(self, dep: "Node", verify_fn: Callable = None) -> bool:
        """
        Check if the given dependency is satisfied. A dependency is satisfied if:
        - a) The upstream has been deployed and no verification function `verify_fn` is
           given for the dependency
        - b) The upstream has been deployed and the given verification
           function `verify_fn` passes

        Parameters:
        dep (Node): The dependency to check.
        verify_fn (Callable, optional): A verification function to call. Defaults to None.

        Returns:
        bool: True if the dependency is satisfied, False otherwise.
        """
        dep_name = dep.get_name()
        if dep not in self._verified_nodes:
            log.debug4("%s not yet verified", dep_name)
            return False

        # Runner initialized with upstream verification disabled.
        if not self._verify_upstream:
            log.debug3("%s verified without checking", dep_name)
            return True

        # Run the custom verification function if provided and return the result.
        if verify_fn is None:
            log.debug4("%s verified with no verify_fn", dep_name)
            return True
        log.debug4("%s calling verify_fn", dep_name)
        satisfied = verify_fn()
        log.debug4("%s verify_fn() -> %s", dep_name, satisfied)
        return satisfied

    def _get_ready_nodes(self) -> List[str]:
        """
        Get the list of nodes whose dependencies (upstream) are satisfied and ready to be executed,

        Returns:
            List[str]: A list of node names that are ready to be executed.
        """
        ready_nodes = []
        for node in [
            n for n in self._get_runnable_nodes() if n not in self._started_nodes
        ]:
            node_name = node.get_name()
            log.debug4("Checking if %s is ready", node_name)
            node_deps = node.get_children()
            satisfied_dependencies = [
                (self._dependency_satisfied(dep, verify_fn), dep)
                for dep, verify_fn in node_deps
            ]

            # If all dependencies are satisfied, the node is ready.
            if all(res[0] for res in satisfied_dependencies):
                ready_nodes.append(node)

            # Otherwise wait for the dependencies (upstream) to be satisfied.
            else:
                log.debug3(
                    "[%s] waiting on upstreams: %s",
                    node_name,
                    [res[0] for res in satisfied_dependencies if not res[0]],
                )
        return ready_nodes

    def _get_completed_nodes(self) -> List[str]:
        return self._verified_nodes + self._unverified_nodes + self._failed_nodes

    def _get_runnable_nodes(self) -> List[Node]:
        return set(self.graph.get_all_nodes()) - set(self._disabled_nodes)


## Helper Classes ##############################################################


class NonThreadPoolExecutor(Executor):
    """This "pool" implements the Executor interfaces, but runs without any
    threads. This is used when running a Runner without cocurrency
    """

    def __init__(self, *_, **__):
        """Swallow constructor args so that it can match ThreadPoolExecutor"""
        super().__init__()

    @staticmethod
    def submit(fn: Callable, /, *args, **kwargs):
        """Run the function immediately and return a pre-completed Future"""
        fut = Future()
        fut.set_result(fn(*args, **kwargs))
        return fut

    @staticmethod
    def shutdown(*_, **__):
        """Nothing to do since this is not a real pool"""


class DagHaltError(Exception):
    """Custom exception used to indicate that a Runner execution should halt"""

    def __init__(
        self,
        failure: bool,
        exception: Exception = None,
    ):
        super().__init__()
        self.failure = failure
        self.exception = exception



"""
Tests of the postgres factory
"""


# Local
from oper8.test_helpers.helpers import setup_session_ctx
from oper8.x.datastores.postgres.factory import PostgresFactory
from tests.x.datastores.postgres.util import (
    POSTGRES_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES,
    set_postgres_secrets,
)

## Helpers #####################################################################


def get_config_overrides(config):
    return {"postgres": config}


## Tests #######################################################################


def test_get_component_with_provided_connection():
    """Test that a provided connection passed in through the CR returns None on a get_component call"""
    override_deploy_configs = POSTGRES_PROVIDED_CONNECTION_DEPLOY_CONFIG_OVERRIDES

    with setup_session_ctx(deploy_config=override_deploy_configs) as session:
        set_postgres_secrets(session)
        component = PostgresFactory.get_component(session)
        # Since the component is provided, we should be getting None back
        assert component is None



"""
Test the DAG Runner functionality
"""

# Standard
import time

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.dag import CompletionState, DagHaltError, Graph, Node, Runner

log = alog.use_channel("TEST")

################################################################################
## Runner Tests #############################################################
################################################################################


class TestRunner:
    """Test the DAG runner directly"""

    #################
    ## Happy Paths ##
    #################

    @pytest.mark.parametrize("threads", [0, 1, 2])
    def test_happy_path(self, threads):
        """Make sure that a simple DAG with dependencies completes and that
        order is correctly preserved
        """

        # Set up nodes in a diamond DAG
        #
        #    A
        #   / \
        #  B  C
        #  \ /
        #   D
        end_order = []

        def node_a():
            end_order.append("A")

        def node_b():
            end_order.append("B")

        def node_c():
            end_order.append("C")

        def node_d():
            end_order.append("D")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node(Node("D", node_d))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("A"))
        graph.add_node_dependency(Node("D"), Node("B"))
        graph.add_node_dependency(Node("D"), Node("C"))

        # Run it
        runner = Runner(threads=threads, graph=graph)
        runner.run()

        # Make sure A came before B and C and D can last
        assert len(end_order) == 4
        assert end_order[0] == "A"
        assert end_order[-1] == "D"
        assert "B" in end_order
        assert "C" in end_order

    def test_multi_root(self):
        """Make sure that a DAG with multiple roots executes to completion"""

        # Set up nodes with multiple roots
        #
        #    A B
        #      |
        #      C
        end_order = []

        def node_a():
            end_order.append("A")

        def node_b():
            end_order.append("B")

        def node_c():
            end_order.append("C")

        graph = Graph()
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("C", node_c))
        graph.add_node_dependency(Node("C"), Node("B"))

        # Run it
        runner = Runner(graph=graph)
        runner.run()

        # Make sure C came after B
        assert len(end_order) == 3
        assert "A" in end_order
        assert "B" in end_order
        assert "C" in end_order
        assert end_order.index("C") > end_order.index("B")

    def test_slow_nodes(self):
        """Make sure that a DAG with slow nodes that have GIL releases will
        still execute in the correct order
        """

        # Set up nodes where one is slow and has a sleep
        #
        #    A
        #   / \
        #  B  C (slow)
        #  |  |
        #  D  E (slow)
        start_order = []
        end_order = []

        def node_a():
            start_order.append("A")
            end_order.append("A")

        def node_b():
            start_order.append("B")
            end_order.append("B")

        def node_c():
            start_order.append("C")
            time.sleep(0.5)
            end_order.append("C")

        def node_d():
            start_order.append("D")
            end_order.append("D")

        def node_e():
            start_order.append("E")
            time.sleep(0.1)
            end_order.append("E")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node(Node("D", node_d))
        graph.add_node(Node("E", node_e))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("A"))
        graph.add_node_dependency(Node("D"), Node("B"))
        graph.add_node_dependency(Node("E"), Node("C"))

        # Run it
        runner = Runner(graph=graph)
        runner.run()
        assert len(start_order) == 5
        assert len(end_order) == 5

        # A start/end before any others
        assert start_order[0] == "A"
        assert end_order[0] == "A"

        # C started before D, but ended after D
        c_start_idx = start_order.index("C")
        c_end_idx = end_order.index("C")
        d_start_idx = start_order.index("D")
        d_end_idx = end_order.index("D")
        assert c_start_idx < d_start_idx
        assert c_end_idx > d_end_idx

        # E start/end after C
        e_start_idx = start_order.index("E")
        e_end_idx = end_order.index("E")
        assert c_start_idx < e_start_idx
        assert c_end_idx < e_end_idx

    def test_incomplete_node(self):
        """Make sure that if a node throws a non-failing DagHaltError, all other
        nodes that can be run are run to completion.
        """

        # Set a graph with an intermediate node that will throw a non-failing
        # DagHaltError
        #
        #    A
        #   / \
        #  B [C]
        #  |  |
        #  D  E
        end_order = []

        def node_a():
            end_order.append("A")

        def node_b():
            end_order.append("B")

        def node_c():
            raise DagHaltError(failure=False)

        def node_d():
            end_order.append("D")

        def node_e():
            end_order.append("E")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node(Node("D", node_d))
        graph.add_node(Node("E", node_e))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("A"))
        graph.add_node_dependency(Node("D"), Node("B"))
        graph.add_node_dependency(Node("E"), Node("C"))

        # Run it
        runner = Runner(graph=graph)
        runner.run()

        # Make sure the completion order looks like A, B, D
        assert end_order == ["A", "B", "D"]

        # Make sure that the runner indicates an incomplete termination state
        completion_state = runner.completion_state()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()

        # Make sure that the completion state looks correct
        assert completion_state == CompletionState(
            verified_nodes=[Node("A"), Node("B"), Node("D")],
            unverified_nodes=[Node("C")],
            failed_nodes=[],
            unstarted_nodes=[Node("E")],
        )

    def test_custom_verify_pass(self):
        """Make sure that giving a dependency a custom verify function which
        passes allows the graph to proceed to a dependent node
        """
        # Set up a simple graph with a custom verify on the one edge
        #
        #    A
        #   [|]
        #    B
        def empty_node():
            pass

        def custom_verify():
            log.debug("Custom verify called!")
            return True

        graph = Graph()
        graph.add_node(Node("A", empty_node))
        graph.add_node(Node("B", empty_node))
        graph.add_node_dependency(Node("B"), Node("A"), custom_verify)

        # Run it
        runner = Runner(graph=graph)
        runner.run()

        # Make sure that the runner indicates full completion
        completion_state = runner.completion_state()
        assert completion_state.verify_completed()
        assert not completion_state.failed()

        # Make sure that the completion state looks correct
        assert completion_state == CompletionState(
            verified_nodes=[Node("A"), Node("B")],
            unverified_nodes=[],
            failed_nodes=[],
            unstarted_nodes=[],
        )

    def test_custom_verify_fail(self):
        """Make sure that giving a dependency a custom verify function which
        fails stops the graph from proceeding to the dependent node
        """

        # Set up a simple graph with a custom verify on the one edge
        #
        #    A
        #   [|]
        #    B
        def empty_node():
            pass

        def custom_verify():
            log.debug("Custom verify called!")
            return False

        graph = Graph()
        graph.add_node(Node("A", empty_node))
        graph.add_node(Node("B", empty_node))
        graph.add_node_dependency(Node("B"), Node("A"), custom_verify)

        # Run it
        runner = Runner(graph=graph)
        runner.run()

        # Make sure that the runner indicates incomplete termination
        completion_state = runner.completion_state()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()

        # Make sure that the completion state looks correct
        assert completion_state == CompletionState(
            verified_nodes=[Node("A")],
            unverified_nodes=[],
            failed_nodes=[],
            unstarted_nodes=[Node("B")],
        )

    def test_custom_verify_both(self):
        """Make sure that two nodes can depend on the same upstream but use
        different verification functions
        """

        # Set up a graph with two nodes off of the root, one of which has a
        # custom verify that passes and the other of which has a custom verify
        # that fails.
        #
        #    A
        #   /[\]
        #  B  C
        def empty_node():
            pass

        def custom_verify_fail():
            log.debug("Custom verify FAIL called!")
            return False

        def custom_verify_pass():
            log.debug("Custom verify PASS called!")
            return True

        graph = Graph()
        graph.add_node(Node("A", empty_node))
        graph.add_node(Node("B", empty_node))
        graph.add_node(Node("C", empty_node))
        graph.add_node_dependency(Node("B"), Node("A"), custom_verify_pass)
        graph.add_node_dependency(Node("C"), Node("A"), custom_verify_fail)

        # Run it
        runner = Runner(graph=graph)
        runner.run()

        # Make sure that the runner indicates incomplete termination
        completion_state = runner.completion_state()
        assert not completion_state.verify_completed()
        assert not completion_state.failed()

        # Make sure that the completion state looks correct
        assert completion_state == CompletionState(
            verified_nodes=[Node("A"), Node("B")],
            unverified_nodes=[],
            failed_nodes=[],
            unstarted_nodes=[Node("C")],
        )

    def test_disable_nodes(self):
        # Set up nodes in a diamond DAG
        #
        #    A
        #   / \
        #  B  C
        #  \ /
        #   D
        end_order = []

        def node_a():
            end_order.append("A")

        def node_b():
            end_order.append("B")

        def node_c():
            end_order.append("C")

        def node_d():
            end_order.append("D")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node(Node("D", node_d))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("A"))
        graph.add_node_dependency(Node("D"), Node("B"))
        graph.add_node_dependency(Node("D"), Node("C"))

        # Disable node D

        # Run it
        runner = Runner(graph=graph)
        runner.disable_node(Node("D"))
        runner.run()

        # Make sure only three nodes ran
        assert len(end_order) == 3
        assert end_order[0] == "A"
        assert "D" not in end_order

    def test_disable_reenable_nodes(self):
        # Set up nodes in a diamond DAG
        #
        #    A
        #   / \
        #  B  C
        #  \ /
        #   D
        end_order = []

        def node_a():
            end_order.append("A")

        def node_b():
            end_order.append("B")

        def node_c():
            end_order.append("C")

        def node_d():
            end_order.append("D")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node(Node("D", node_d))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("A"))
        graph.add_node_dependency(Node("D"), Node("B"))
        graph.add_node_dependency(Node("D"), Node("C"))

        # Disable node D

        # Run it
        runner = Runner(graph=graph)
        runner.disable_node(Node("D"))
        runner.enable_node(Node("D"))
        runner.run()

        assert len(end_order) == 4
        assert end_order[0] == "A"
        assert "D" in end_order

    #################
    ## Error Cases ##
    #################

    def test_node_throw_dag(self):
        """Make sure that if a node raises a DagHaltError, the DAG nodes before
        it still completed but nodes after it do not
        """

        # Set up linear nodes with a throw in the middle
        #
        #    A -> B -x C
        start_order = []
        end_order = []

        def node_a():
            start_order.append("A")
            end_order.append("A")

        def node_b():
            start_order.append("B")
            raise DagHaltError()

        def node_c():
            start_order.append("C")
            end_order.append("C")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("B"))

        # Run it
        runner = Runner(graph=graph)
        runner.run()
        assert start_order == ["A", "B"]
        assert end_order == ["A"]

    def test_node_throw_runtime(self):
        """Make sure that if a node throws, the DAG nodes before it still
        completed but nodes after it do not
        """

        # Set up linear nodes with a throw in the middle
        #
        #    A -> B -x C
        start_order = []
        end_order = []

        def node_a():
            start_order.append("A")
            end_order.append("A")

        def node_b():
            start_order.append("B")
            raise RuntimeError("Nope!")

        def node_c():
            start_order.append("C")
            end_order.append("C")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))
        graph.add_node(Node("C", node_c))
        graph.add_node_dependency(Node("B"), Node("A"))
        graph.add_node_dependency(Node("C"), Node("B"))

        # Run it
        runner = Runner(graph=graph)
        runner.run()
        assert start_order == ["A", "B"]
        assert end_order == ["A"]

    def test_node_in_flight(self):
        """Make sure that an in-flight node completes when another node causes a
        halt in execution.
        """

        # Set up two independent nodes, one of which will terminate and the
        # the other of which takes a while
        start_order = []
        end_order = []

        def node_a():
            start_order.append("A")
            time.sleep(0.1)
            end_order.append("A")

        def node_b():
            time.sleep(0.01)
            start_order.append("B")
            raise RuntimeError("Nope!")

        graph = Graph()
        graph.add_node(Node("A", node_a))
        graph.add_node(Node("B", node_b))

        # Run it and make sure A makes it to the end
        runner = Runner(graph=graph)
        runner.run()
        assert start_order == ["A", "B"]
        assert end_order == ["A"]

    def test_dag_early_halt(self):
        """This test exercises a nasty corner case that came up in WA when the
        CLU subsystem introduced custom verify_functions for all components. The
        bug is:

        With slow verification functions, a node running in a thread
        may complete and mark itself verified after a downstream dependency has
        checked its completion status, but before the full set of
        _get_ready_nodes() checks has passed. If this happens and
        _get_completed_nodes() is called afterwards, the short-circuit logic
        below will think that all started nodes have completed and there are no
        ready nodes, thus terminating the DAG prematurely.
        """

        class NodeTest(Node):
            def __init__(
                self, name, call_sleep=0.1, verify_sleep=0.1, verify_fail_times=0
            ):
                super().__init__(name, self)
                self.name = name
                self.call_sleep = call_sleep
                self.verify_sleep = verify_sleep
                self.verify_fail_times = verify_fail_times
                self.verify_call = 0

            def __repr__(self):
                return f"Node[{self.name}]"

            def __call__(self):
                log.info("Calling %s", self)
                time.sleep(self.call_sleep)
                log.info("Done calling %s", self)

            def verify(self):
                time.sleep(self.verify_sleep)
                if self.verify_call < self.verify_fail_times:
                    self.verify_call += 1
                    return False
                return True

        slow_call_parent = NodeTest("A", call_sleep=0.1, verify_sleep=0)
        slow_verify_parent = NodeTest(
            "B", call_sleep=0, verify_sleep=0.1, verify_fail_times=1
        )
        child_c = NodeTest("C")
        child_d = NodeTest("D")

        graph = Graph()
        graph.add_node(slow_call_parent)
        graph.add_node(slow_verify_parent)
        graph.add_node(child_c)
        graph.add_node(child_d)
        graph.add_node_dependency(child_c, slow_call_parent, slow_call_parent.verify)
        graph.add_node_dependency(
            child_d, slow_verify_parent, slow_verify_parent.verify
        )
        runner = Runner("test", graph=graph)
        runner.run()
        completion_state = runner.completion_state()
        assert completion_state.verify_completed()



"""
The ReconcileManager class manages an individual reconcile of a controller.
This setups up the session, constructs the controller, and runs its reconcile
"""

# Standard
from dataclasses import dataclass, field
from typing import List, Optional, Set, Type, Union
import base64
import copy
import datetime
import importlib
import json
import logging
import os
import pathlib
import random
import re
import sys
import uuid

# Third Party
import dateutil

# First Party
import aconfig
import alog

# Local
from . import config, constants, status
from .dag import CompletionState
from .deploy_manager import (
    DeployManagerBase,
    DryRunDeployManager,
    OpenshiftDeployManager,
)
from .exceptions import (
    ClusterError,
    ConfigError,
    Oper8ExpectedError,
    Oper8FatalError,
    PreconditionError,
    RolloutError,
    VerificationError,
    assert_cluster,
    assert_config,
)
from .log_format import Oper8JsonFormatter
from .session import Session
from .utils import add_finalizer, get_manifest_version, merge_configs, remove_finalizer
from .vcs import VCS, VCSCheckoutMethod, VCSMultiProcessError

log = alog.use_channel("RECONCILE")


## Data models #################################################################


@dataclass
class RequeueParams:
    """RequeueParams holds parameters for requeue request"""

    requeue_after: datetime.timedelta = field(
        default_factory=lambda: datetime.timedelta(
            seconds=float(config.requeue_after_seconds)
        )
    )


@dataclass
class ReconciliationResult:
    """ReconciliationResult is the result of a reconciliation session"""

    # Flag to control requeue of current reconcile request
    requeue: bool
    # Parameters for requeue request
    requeue_params: RequeueParams = field(default_factory=RequeueParams)
    # Flag to identify if the reconciliation raised an exception
    exception: Exception = None


# Forward declarations of Controller
CONTROLLER_TYPE = "Controller"
CONTROLLER_CLASS_TYPE = Type[CONTROLLER_TYPE]

# Type helper for describing a controller. CONTROLLER_INFO
# can be a str in the form "module.class", a Controller class that
# will be initialized, or an already initialized Controller. While the
# first two methods are preferred, an already created Controller can
# be useful for tests and backwards compatibility
CONTROLLER_INFO = Union[str, CONTROLLER_CLASS_TYPE, CONTROLLER_TYPE]

## ReconcileManager #################################################################


class ReconcileManager:  # pylint: disable=too-many-lines
    """This class manages reconciliations for an instance of Oper8. It's
    primary function is to run reconciles given a CR manifest, Controller,
    and the current cluster state via a DeployManager.
    """

    ## Construction ############################################################

    def __init__(
        self,
        home_dir: str = None,
        deploy_manager: Optional[DeployManagerBase] = None,
        enable_vcs: Optional[bool] = None,
        reimport_controller: Optional[bool] = True,
    ):
        """The constructor sets up the properties used across every
        reconcile and checks that the current config is valid.

        Args:
            home_dir:  Optional[str]=None
                The root directory for importing controllers or VCS checkout
            deploy_manager:  Optional[DeployManager]=None
                Deploy manager to use. If not given, a new DeployManager will
                be created for each reconcile.
            enable_vcs:  Optional[bool]=True
                Parameter to manually control the state of VCS on a per instance
                basis
            reimport_controller:  Optional[bool]=None
                Parameter to manually control if a controller needs to be reimported each
                reconcile.
        """

        if home_dir:
            self.home_dir = home_dir
        elif config.vcs.enabled:
            self.home_dir = config.vcs.repo
        else:
            self.home_dir = os.getcwd()

        self.vcs = None

        # If enable_vcs is not provided than default to
        # config
        if enable_vcs is None:
            enable_vcs = config.vcs.enabled

        if enable_vcs:
            assert_config(
                config.vcs.repo,
                "Can not enable vcs without supply source repo at vcs.repo",
            )
            assert_config(
                config.vcs.dest,
                "Cannot require enable vcs without providing a destination",
            )
            vcs_checkout_methods = [method.value for method in VCSCheckoutMethod]
            assert_config(
                config.vcs.checkout_method in vcs_checkout_methods,
                f"VCS checkout method must be one of the following {vcs_checkout_methods}",
            )

            self.vcs = VCS(self.home_dir)

        # Ensure config is setup correctly for strict_versioning
        self.strict_version_regex = re.compile(config.strict_version_kind_regex or ".*")
        if config.strict_versioning:
            assert_config(
                config.supported_versions is not None,
                "Must provide supported_versions with strict_versioning=True",
            )
            assert_config(
                config.vcs.field is not None,
                "Must provide vcs.field with strict_versioning=True",
            )

        self.vcs_kind_regex = re.compile(config.vcs.kind_regex or ".*")

        self.deploy_manager = deploy_manager
        self.reimport_controller = reimport_controller

    ## Reconciliation ############################################################

    @alog.logged_function(log.info)
    @alog.timed_function(log.info, "Reconcile finished in: ")
    def reconcile(
        self,
        controller_info: CONTROLLER_INFO,
        resource: Union[dict, aconfig.Config],
        is_finalizer: bool = False,
    ) -> ReconciliationResult:
        """This is the main entrypoint for reconciliations and contains the
        core implementation. The general reconcile path is as follows:

            1. Parse the raw CR manifest
            2. Setup logging based on config with overrides from CR
            3. Check if the CR is paused and for strict versioning
            4. Setup directory if VCS is enabled
            5. Import and construct the Controller
            6. Setup the DeployManager and Session objects
            7. Run the Controller reconcile

        Args:
            controller_info: CONTROLLER_INFO
                The description of a controller. See CONTROLLER_INFO for
                more information
            resource: Union[dict, aconfig.Config]
                A raw representation of the resource to be reconciled
            is_finalizer: bool=False
                Whether the resource is being deleted

        Returns:
            reconcile_result:  ReconciliationResult
                The result of the reconcile
        """

        # Parse the full CR content
        cr_manifest = self.parse_manifest(resource)

        # generate id unique to this session
        reconcile_id = self.generate_id()

        # Initialize logging prior to any other work
        self.configure_logging(cr_manifest, reconcile_id)

        # If paused, do nothing and don't requeue
        if self._is_paused(cr_manifest):
            log.info("CR is paused. Exiting reconciliation")
            result = ReconciliationResult(requeue=False, requeue_params=RequeueParams())
            return result

        # Check strict versioning before continuing
        if self._should_check_strict_version(cr_manifest):
            self._check_strict_versioning(cr_manifest)

        # Check if VCS is enabled and then attempt to checkout
        if config.vcs.enabled:
            self.setup_vcs(cr_manifest)

        # Import controller and setup the instance
        controller = self.setup_controller(controller_info)

        # Configure deploy manager on a per reconcile basis for
        # owner references unless a manager is provided on initialization
        deploy_manager = self.setup_deploy_manager(cr_manifest)

        # Setup Session
        session = self.setup_session(
            controller, cr_manifest, deploy_manager, reconcile_id
        )

        # Run the controller reconcile
        result = self.run_controller(controller, session, is_finalizer)

        return result

    def safe_reconcile(
        self,
        controller_info: CONTROLLER_INFO,
        resource: dict,
        is_finalizer: bool = False,
    ) -> ReconciliationResult:
        """
        This function calls out to reconcile but catches any errors thrown. This
        function guarantees a safe result which is needed by some Watch Managers

        Args:
            controller_info: CONTROLLER_INFO
                The description of a controller. See CONTROLLER_INFO for
                more information
            resource: Union[dict, aconfig.Config]
                A raw representation of the reconcile
            is_finalize: bool=False
                Whether the resource is being deleted

        Returns:
            reconcile_result:  ReconciliationResult
                The result of the reconcile

        """

        try:
            return self.reconcile(controller_info, resource, is_finalizer)

        # VCSMultiProcessError is an expected error caused by oper8 which should
        # not be handled by the exception handling code
        except VCSMultiProcessError as exc:
            # Requeue after ~7.5 seconds. Add randomness to avoid
            # repeated conflicts
            requeue_time = 5 + random.uniform(0, 5)
            params = RequeueParams(
                requeue_after=datetime.timedelta(seconds=requeue_time)
            )
            log.debug("VCS Multiprocessing Error Detected: {%s}", exc, exc_info=True)
            log.warning(
                "VCS Setup failed due to other process. Requeueing in %ss",
                requeue_time,
            )
            return ReconciliationResult(
                requeue=True, requeue_params=params, exception=exc
            )

        # Capture all generic exceptions
        except Exception as exc:  # pylint: disable=broad-except
            log.warning("Handling caught error in reconcile: %s", exc, exc_info=True)
            error = exc

        if config.manage_status:
            try:
                self._update_error_status(resource, error)
                log.debug("Update CR status with error message")
            except Exception as exc:  # pylint: disable=broad-except
                log.error("Failed to update status: %s", exc, exc_info=True)

        # If we got to this return it means there was an
        # exception during reconcile and we should requeue
        # with the default backoff period
        log.info("Requeuing CR due to error during reconcile")
        return ReconciliationResult(
            requeue=True, requeue_params=RequeueParams(), exception=error
        )

    ## Reconciliation Stages ############################################################

    @classmethod
    def parse_manifest(cls, resource: Union[dict, aconfig.Config]) -> aconfig.Config:
        """Parse a raw resource into an aconfig Config

        Args:
            resource: Union[dict, aconfig.Config])
                The resource to be parsed into a manifest

        Returns
            cr_manifest: aconfig.Config
                The parsed and validated config
        """
        try:
            cr_manifest = aconfig.Config(resource, override_env_vars=False)
        except (ValueError, SyntaxError, AttributeError) as exc:
            raise ValueError("Failed to parse full_cr") from exc

        return cr_manifest

    @classmethod
    def configure_logging(cls, cr_manifest: aconfig.Config, reconciliation_id: str):
        """Configure the logging for a given reconcile

        Args:
            cr_manifest: aconfig.Config
                The resource to get annotation overrides from
            reconciliation_id: str
                The unique id for the reconciliation
        """

        # Fetch the annotations for logging
        # NOTE: We use safe fetching here because this happens before CR
        #   verification in the Session constructor
        annotations = cr_manifest.get("metadata", {}).get("annotations", {})
        default_level = annotations.get(
            constants.LOG_DEFAULT_LEVEL_NAME, config.log_level
        )

        filters = annotations.get(constants.LOG_FILTERS_NAME, config.log_filters)
        log_json = annotations.get(constants.LOG_JSON_NAME, str(config.log_json))
        log_thread_id = annotations.get(
            constants.LOG_THREAD_ID_NAME, str(config.log_thread_id)
        )

        # Convert boolean args
        log_json = (log_json or "").lower() == "true"
        log_thread_id = (log_thread_id or "").lower() == "true"

        # Keep the old handler. This is useful if running with ansible as
        # it will preserve the handler generator set up to log to a file
        # since ansible captures all logging output
        handler_generator = None
        if logging.root.handlers:
            old_handler = logging.root.handlers[0]

            def handler_generator():
                return old_handler

        alog.configure(
            default_level=default_level,
            filters=filters,
            formatter=(
                Oper8JsonFormatter(cr_manifest, reconciliation_id)
                if log_json
                else "pretty"
            ),
            thread_id=log_thread_id,
            handler_generator=handler_generator,
        )

    @classmethod
    def generate_id(cls) -> str:
        """Generates a unique human readable id for this reconciliation

        Returns:
            id: str
                A unique base32 encoded id
        """
        uuid4 = uuid.uuid4()
        base32_str = base64.b32encode(uuid4.bytes).decode("utf-8")
        reconcile_id = base32_str[:22]
        log.debug("Generated reconcile id: %s", reconcile_id)
        return reconcile_id

    def setup_vcs(self, cr_manifest: aconfig.Config):
        """Setups the VCS directory and sys.path for a reconcile.
        This function also ensures that the version is valid if
        config.strict_versioning is enabled.

        Args:
            cr_manifest: aconfig.Config
                The cr manifest to pull the requested version from.
        """
        if not self.vcs_kind_regex.match(cr_manifest.get("kind", "")):
            log.debug("Skipping vcs checkout due to kind regex")
            return

        version = get_manifest_version(cr_manifest)
        if not version:
            raise ValueError("CR Manifest has no version")

        log.debug(
            "Setting up working directory with src: %s and version: %s",
            self.home_dir,
            version,
        )
        working_dir = self._setup_directory(cr_manifest, version)

        # Construct working dir path from vcs and git directory
        if config.vcs.module_dir:
            module_path = pathlib.Path(config.vcs.module_dir)
            working_dir = working_dir / module_path

        if not working_dir.is_dir():
            log.error(
                "Working directory %s could not be found. Invalid module path",
                working_dir,
            )
            raise ConfigError(
                f"Module path: '{module_path}' could not be found in repository"
            )

        log.debug4("Changing working directory to %s", working_dir)
        os.chdir(working_dir)
        sys.path.insert(0, str(working_dir))

    def setup_controller(
        self, controller_info: CONTROLLER_INFO
    ) -> CONTROLLER_CLASS_TYPE:
        """
        Import the requested Controller class and enable any compatibility layers

        Args:
            controller_info:CONTROLLER_INFO
                The description of a controller. See CONTROLLER_INFO for
                more information
        Returns:
            controller:
                The required Controller Class
        """

        # Local
        from .controller import (  # pylint: disable=import-outside-toplevel, cyclic-import
            Controller,
        )

        # If controller info is already a constructed controller then
        # skip importing
        if isinstance(controller_info, Controller):
            return controller_info

        controller_class = self._import_controller(controller_info)
        return self._configure_controller(controller_class)

    def setup_deploy_manager(self, cr_manifest: aconfig.Config) -> DeployManagerBase:
        """
        Configure a deploy_manager for a reconcile given a manifest

        Args:
            cr_manifest: aconfig.Config
                The resource to be used as an owner_ref

        Returns:
            deploy_manager: DeployManagerBase
                The deploy_manager to be used during reconcile
        """
        if self.deploy_manager:
            return self.deploy_manager

        if config.dry_run:
            log.debug("Using DryRunDeployManager")
            return DryRunDeployManager()

        log.debug("Using OpenshiftDeployManager")
        return OpenshiftDeployManager(owner_cr=cr_manifest)

    def setup_session(
        self,
        controller: CONTROLLER_TYPE,
        cr_manifest: aconfig.Config,
        deploy_manager: DeployManagerBase,
        reconciliation_id: str,
    ) -> Session:
        """Construct the session, including gathering the backend config and any temp patches

        Args:
            controller: Controller
                The controller class being reconciled
            cr_manifest: aconfig.Config
                The resource being reconciled
            deploy_manager: DeployManagerBase
                The deploy manager used in the cluster
            reconciliation_id: str
                The id for the reconcile

        Return:
            session: Session
                The session for reconcile
        """
        # Get backend config for reconciliation
        controller_defaults = controller.get_config_defaults()
        reconciliation_config = self._get_reconcile_config(
            cr_manifest=cr_manifest,
            deploy_manager=deploy_manager,
            controller_defaults=controller_defaults,
        )
        log.debug4("Gathered Config: %s", reconciliation_config)

        # Get Temporary patches
        patches = self._get_temp_patches(deploy_manager, cr_manifest)
        log.debug3("Found %d patches", len(patches))

        # Get the complete CR Manifest including defaults
        cr_manifest_defaults = controller.get_cr_manifest_defaults()
        full_cr_manifest = merge_configs(
            aconfig.Config(cr_manifest_defaults),
            cr_manifest,
        )

        return Session(
            reconciliation_id=reconciliation_id,
            cr_manifest=full_cr_manifest,
            config=reconciliation_config,
            deploy_manager=deploy_manager,
            temporary_patches=patches,
        )

    def run_controller(
        self, controller: CONTROLLER_TYPE, session: Session, is_finalizer: bool
    ) -> ReconciliationResult:
        """Run the Controller's reconciliation or finalizer with the constructed Session.
        This function also updates the CR status and handles requeue logic.

        Args:
            controller: Controller
                The Controller being reconciled
            session: Session
                The current Session state
            is_finalizer:
                Whether the resource is being deleted

        Returns:
            reconciliation_result: ReconciliationResult
                The result of the reconcile
        """
        log.info(
            "%s resource %s/%s/%s",
            "Finalizing" if is_finalizer else "Reconciling",
            session.kind,
            session.namespace,
            session.name,
        )

        # Ensure the resource has the proper finalizers
        if controller.has_finalizer:
            add_finalizer(session, controller.finalizer)

        # Update the Resource status
        if config.manage_status:
            self._update_reconcile_start_status(session)

        # Reconcile the controller
        completion_state = controller.run_reconcile(
            session,
            is_finalizer=is_finalizer,
        )

        if config.manage_status:
            self._update_reconcile_completion_status(session, completion_state)

        # Check if the controller session should requeue
        requeue, requeue_params = controller.should_requeue(session)
        if not requeue_params:
            requeue_params = RequeueParams()

        # Remove managed finalizers if not requeuing
        if not requeue and is_finalizer and controller.has_finalizer:
            remove_finalizer(session, controller.finalizer)

        return ReconciliationResult(requeue=requeue, requeue_params=requeue_params)

    ## Implementation Details ############################################################

    @classmethod
    def _is_paused(cls, cr_manifest: aconfig.Config) -> bool:
        """Check if a manifest has a paused annotation

        Args:
            cr_manifest: aconfig.Config
                The manifest becking checked

        Returns:
            is_paused: bool
                If the manifest contains the paused annotation
        """
        annotations = cr_manifest.metadata.get("annotations", {})
        paused = annotations.get(constants.PAUSE_ANNOTATION_NAME)
        return paused and paused.lower() == "true"

    def _check_strict_versioning(self, cr_manifest: aconfig.Config):
        """Check the version against config and vcs directory

        Args:
            version_directory: str
                The repo directory to check
            version: str
                The version from the manifest
        """
        version = get_manifest_version(cr_manifest)
        if not version:
            raise ValueError("CR Manifest has no version")

        # Ensure version is in list of supported versions
        assert_config(
            version in config.supported_versions,
            f"Unsupported version: {version}",
        )

        # If VCS is enabled ensure the branch or tag exists
        if self.vcs:
            repo_versions = self.vcs.list_refs()
            assert_config(
                version in repo_versions,
                f"Version not found in repo: {version}",
            )
            log.debug3("Supported VCS Versions: %s", repo_versions)

    def _should_check_strict_version(self, cr_manifest: aconfig.Config) -> bool:
        """Helper function to check if we should enforce strict versioning. Checks the
        general config and then matches the kind regex

        Args:
            cr_manifest: aconfig.Config
                The CR for this reconciliation

        Returns:
            bool: True if we should check this kind
        """
        if not config.strict_versioning:
            return False

        return bool(self.strict_version_regex.match(cr_manifest.get("kind")))

    def _setup_directory(
        self, cr_manifest: aconfig.Config, version: str
    ) -> pathlib.Path:
        """Construct the VCS directory from the cr_manifest and version. Then
        checkout the directory

        Args:
            cr_manifest: aconfig.Config
                The manifest to be used for the checkout path
            version: str
                The version to checkout

        Returns:
            destination_directory: pathlib.Path
                The destination directory for the checkout
        """

        # Generate checkout directory and ensure path exists
        def sanitize_for_path(path):
            keepcharacters = (" ", ".", "_")
            return "".join(
                c for c in path if c.isalnum() or c in keepcharacters
            ).rstrip()

        # Setup destination template to allow for CR specific checkout paths
        # The entirety of the cr_manifest is included as a dict as well as some
        # custom keys
        template_mappings = {
            # Include the entire dict first so that the sanitized default values
            # take precedence
            **cr_manifest,
            "version": version,
            "kind": sanitize_for_path(cr_manifest.kind),
            "apiVersion": sanitize_for_path(cr_manifest.apiVersion),
            "namespace": sanitize_for_path(cr_manifest.metadata.namespace),
            "name": sanitize_for_path(cr_manifest.metadata.name),
        }

        # Get the checkout directory and method
        try:
            formatted_path = config.vcs.dest.format(**template_mappings)
        except KeyError as exc:
            log.warning(
                "Invalid key: %s found in vcs destination template", exc, exc_info=True
            )
            raise ConfigError("Invalid Key found in vcs destination template") from exc

        checkout_dir = pathlib.Path(formatted_path)
        checkout_method = VCSCheckoutMethod(config.vcs.checkout_method)

        log.debug2(
            "Checking out into directory %s with method %s",
            checkout_dir,
            checkout_method.value,
        )
        self.vcs.checkout_ref(version, checkout_dir, checkout_method)
        return checkout_dir

    @staticmethod
    def _unimport_controller_module(module_name: str) -> Set[str]:
        """Helper to un-import the given module and its parents/siblings/
        children

        Args:
            module_name: str
                The name of the module that holds the Controller

        Returns:
            reimport_modules: Set[str]
                All modules that were unimported and will need to be reimported
        """
        reimport_modules = set()
        if module_name in sys.modules:
            log.debug2("UnImporting controller module: %s", module_name)
            sys.modules.pop(module_name)
            reimport_modules.add(module_name)

        # UnImport the controller and any parent/sibling/child modules so
        # controller can be reimported from the most recent sys path
        module_parts = module_name.split(".")
        for i in range(1, len(module_parts)):
            parent_module = ".".join(module_parts[:-i])
            if parent_module in sys.modules:
                log.debug3("UnImporting module: %s", parent_module)
                if sys.modules.pop(parent_module, None):
                    reimport_modules.add(parent_module)
        for child_module in [
            mod_name
            for mod_name in sys.modules
            if mod_name.startswith(f"{module_parts[0]}.")
        ]:
            log.debug3("UnImporting child module: %s", child_module)
            if sys.modules.pop(child_module, None):
                reimport_modules.add(child_module)
        return reimport_modules

    def _import_controller(
        self, controller_info: CONTROLLER_INFO
    ) -> CONTROLLER_CLASS_TYPE:
        """Parse the controller info and reimport the controller

        Args:
            controller_info:CONTROLLER_INFO
                The description of a controller. See CONTROLLER_INFO for
                more information
        Returns:
            controller_class: Type[Controller]
                The reimported Controller

        """
        log.debug2("Parsing controller_info")
        if isinstance(controller_info, str):
            class_module_parts = controller_info.rsplit(".", maxsplit=1)
            assert_config(
                len(class_module_parts) == 2,
                f"Invalid controller_class [{controller_info}]. Format is <module>.<class>",
            )
            module_name, class_name = class_module_parts
        else:
            class_name = controller_info.__name__
            module_name = controller_info.__module__

        # Reimport module if reimporting is enabled and if it already exists
        log.debug3(
            "Running controller %s from module %s [reimport? %s, in sys.modules? %s]",
            class_name,
            module_name,
            self.reimport_controller,
            module_name in sys.modules,
        )
        reimport_modules = {module_name}
        if self.reimport_controller:
            reimport_modules = reimport_modules.union(
                self._unimport_controller_module(module_name)
            )

        # Attempt to import the modules
        log.debug2("Attempting to import [%s.%s]", module_name, class_name)
        for reimport_name in reimport_modules:
            try:
                app_module = importlib.import_module(reimport_name)
                if reimport_name == module_name:
                    if not hasattr(app_module, class_name):
                        raise ConfigError(
                            f"Invalid controller_class [{class_name}]."
                            f" Class not found in module [{reimport_name}]"
                        )
                    controller_class = getattr(app_module, class_name)

                    # Import controller in function to avoid circular imports
                    # Local
                    from .controller import (  # pylint: disable=import-outside-toplevel
                        Controller,
                    )

                    if not issubclass(controller_class, Controller):
                        raise ConfigError(
                            f"Invalid controller_class [{module_name}.{class_name}]."
                            f" [{class_name}] is not a Controller"
                        )

            except ImportError as exc:
                # If this is the module that holds the controller, it _needs_ to
                # be reimported
                if reimport_name == module_name:
                    log.error(
                        "Failed to import [%s.%s]. Failed to import [%s]",
                        reimport_name,
                        class_name,
                        reimport_name,
                        exc_info=True,
                    )
                    raise ConfigError("Invalid Controller Class Specified") from exc
                # Otherwise, it's ok for import to fail
                else:
                    log.debug("Not able to reimport %s", reimport_name)

        log.debug(
            "Imported Controller %s from file %s",
            controller_class,
            sys.modules[controller_class.__module__].__file__,
        )

        return controller_class

    def _configure_controller(
        self, controller_class: CONTROLLER_CLASS_TYPE
    ) -> CONTROLLER_TYPE:
        """Construct the Controller Class

        Args:
            controller_class: CONTROLLER_CLASS_TYPE
                The Controller class to be reconciled

        Returns:
            controller: Controller
                The constructed Controller

        """
        log.debug3("Constructing controller")
        controller = controller_class()
        return controller

    def _get_reconcile_config(
        self,
        cr_manifest: aconfig.Config,
        deploy_manager: DeployManagerBase,
        controller_defaults: aconfig.Config,
    ) -> aconfig.Config:
        """Construct the flattened backend config for this reconciliation
        starting with a deepcopy of the base and merge in overrides from the CR

        Args:
            cr_manifest: aconfig.Config:
                The manifest to get overrides from
            deploy_manager: DeployManagerBase:
                The deploy manager to get the default configmap config
            controller_defaults: aconfig.Config:
                The config defaults provided by the controller class

        Returns:
            reconcile_config: aconfig.Config
                The reconciliation config
        """
        metadata = cr_manifest.get("metadata", {})
        annotations = metadata.get("annotations", {})
        namespace = metadata.get("namespace")
        cr_config_defaults = cr_manifest.get(constants.CONFIG_OVERRIDES, {})
        annotation_config_defaults = {}
        if constants.CONFIG_DEFAULTS_ANNOTATION_NAME in annotations:
            log.debug("Pulling config_defaults based on annotation")
            config_defaults_name = annotations[
                constants.CONFIG_DEFAULTS_ANNOTATION_NAME
            ]

            # Allow sub-keys to be delineated by "/"
            parts = config_defaults_name.split("/")
            config_defaults_cm_name = parts[0]

            log.debug2(
                "Reading config_defaults from ConfigMap [%s]", config_defaults_cm_name
            )
            success, content = deploy_manager.get_object_current_state(
                kind="ConfigMap",
                name=config_defaults_cm_name,
                namespace=namespace,
                api_version="v1",
            )
            assert_cluster(success, "Failed to look up config defaults form ConfigMap")
            assert_config(
                content is not None,
                f"Did not find configured config defaults ConfigMap: {config_defaults_cm_name}",
            )
            assert_config("data" in content, "Got ConfigMap content with out 'data'")
            config_defaults_content = content["data"]
            assert_config(
                isinstance(config_defaults_content, dict),
                f"Incorrectly formatted config_defaults ConfigMap: {config_defaults_cm_name}",
            )

            # Parse as a Config
            log.debug2("Parsing app config dict")
            annotation_config_defaults = aconfig.Config(
                config_defaults_content, override_env_vars=False
            )

        return merge_configs(
            copy.deepcopy(controller_defaults),
            merge_configs(annotation_config_defaults, cr_config_defaults),
        )

    def _get_temp_patches(  # pylint: disable=too-many-locals
        self, deploy_manager: DeployManagerBase, cr_manifest: aconfig.Config
    ) -> List[aconfig.Config]:
        """Fetch the ordered list of temporary patches that should apply to this
        rollout.

        Args:
            deploy_manager: DeployManagerBase
                The DeployManager used to get the current temporary patches
            cr_manifest: aconfig.Config
                The manifest of this reconciliation
        """

        # Look for patch annotations on the CR
        patch_annotation = (
            cr_manifest.get("metadata", {})
            .get("annotations", {})
            .get(constants.TEMPORARY_PATCHES_ANNOTATION_NAME, "{}")
        )
        log.debug3("Raw patch annotation: %s", patch_annotation)
        try:
            raw_patches = json.loads(patch_annotation)
            if not isinstance(raw_patches, dict):
                msg = f"Patches annotation not a dict: {raw_patches}"
                log.error(msg)
                raise RolloutError(msg)
            patches = {}
            for patch_name, patch_meta in raw_patches.items():
                patch_meta["timestamp"] = dateutil.parser.parse(patch_meta["timestamp"])
                patches[patch_name] = patch_meta
                if "api_version" not in patch_meta:
                    raise KeyError("api_version")
        except json.decoder.JSONDecodeError as err:
            msg = f"Could not parse patches from annotation [{patch_annotation}]"
            log.error(msg)
            raise RolloutError(msg) from err
        except dateutil.parser.ParserError as err:
            msg = f"Failed to parse patch timestamp [{patch_annotation}]"
            log.error(msg)
            raise RolloutError(msg) from err
        except KeyError as err:
            msg = f"Patch meta incorrectly formatted [{patch_annotation}]"
            log.error(msg)
            raise RolloutError(msg) from err

        # Fetch the state of each patch and add it to the output, sorted by
        # timestamp with the earliest first
        temporary_patches = []
        for patch_name, patch_meta in sorted(
            list(patches.items()), key=lambda x: x[1]["timestamp"]
        ):
            # Do the fetch
            log.debug2("Fetching patch [%s/%s]", patch_name, patch_meta["timestamp"])
            namespace = cr_manifest.get("metadata", {}).get("namespace")
            patch_api_version = patch_meta["api_version"]
            patch_kind = patch_meta.get("kind", "TemporaryPatch")
            success, content = deploy_manager.get_object_current_state(
                kind=patch_kind,
                name=patch_name,
                api_version=patch_api_version,
                namespace=namespace,
            )
            assert_cluster(success, f"Failed to fetch patch content for [{patch_name}]")
            assert_config(content is not None, f"Patch not found [{patch_name}]")

            # Pull the patch spec and add it to the list
            assert_config(
                content.get("spec") is not None,
                f"No spec found in patch [{patch_name}]",
            )
            temporary_patches.append(aconfig.Config(content, override_env_vars=False))

        return temporary_patches

    ## Status Details ############################################################

    def _update_resource_status(
        self, deploy_manager: DeployManagerBase, manifest: aconfig.Config, **kwargs
    ) -> dict:
        """Helper function to update the status of a resource given a deploy_manager, manifest
        and status kwargs

        Args:
            deploy_manager: DeployManagerBase
                The DeployManager used to update the resource
            manifest: aconfig.Config
                The manifest of the resource being updated
            **kwargs:
                The key word arguments passed to update_resource_status

        Returns:
            updated_status: dict
                The updated status applied to the resource
        """
        return status.update_resource_status(
            deploy_manager,
            manifest.kind,
            manifest.apiVersion,
            manifest.metadata.name,
            manifest.metadata.namespace,
            **kwargs,
        )

    def _update_reconcile_start_status(self, session: Session):
        """Update the status for a resource at the start of a reconciliation

        Args:
            session: Session
                The session of the reconcile which includes the DeployManager and resource

        """
        ready_condition = status.get_condition(status.READY_CONDITION, session.status)
        ready_reason = ready_condition.get("reason")
        if ready_reason is None or session.current_version is None:
            ready_reason = status.ReadyReason.INITIALIZING

        optional_kwargs = {}
        if session.current_version and session.version != session.current_version:
            log.debug(
                "Version change detected: %s -> %s",
                session.current_version,
                session.version,
            )
            optional_kwargs = {
                "updating_reason": status.UpdatingReason.VERSION_CHANGE,
                "updating_message": "Version Change Started: "
                f"[{session.current_version}] -> [{session.version}]",
            }
            ready_reason = status.ReadyReason.IN_PROGRESS

        self._update_resource_status(
            session.deploy_manager,
            session.cr_manifest,
            ready_reason=ready_reason,
            ready_message=ready_condition.get("message", "Initial Rollout Started"),
            supported_versions=config.supported_versions,
            **optional_kwargs,
        )

    def _update_reconcile_completion_status(
        self, session: Session, completion_state: CompletionState
    ):
        """Perform CR status updates based on the results of the rollout steps. The status logic is
        as follows:
          1. Initial Rollout: Ready-INITIALIZING, Updating-VERIFY_WAIT
          2. Everything complete: Ready-STABLE, Updating-STABLE
          3. Everything except after_verify: Ready-IN_PROGRESS, Updating-STABLE
          4. other: Updating-VERIFY_WAIT

          Args:
            session: Session
                The session of the reconcile which includes the DeployManager and resource
            completion_state: CompletionState
                The result of the rollout
        """
        status_update = {"component_state": completion_state}

        # If everything completed and verified, set ready and updating to STABLE
        # and set the status's reconciled version to the desired version
        if completion_state.verify_completed():
            status_update["ready_reason"] = status.ReadyReason.STABLE
            status_update["ready_message"] = "Verify Complete"
            status_update["updating_reason"] = status.UpdatingReason.STABLE
            status_update["updating_message"] = "Rollout Complete"
            status_update["version"] = session.version

        # If the completion_state didn't fail then update the ready condition with
        # in_progress and the updating condition with verification incomplete
        else:
            current_status = session.get_status()

            # If not initializing then update the ready condition with in_progress
            current_ready_cond = status.get_condition(
                status.READY_CONDITION, current_status
            )
            if (
                current_ready_cond.get("reason")
                != status.ReadyReason.INITIALIZING.value
            ):
                status_update["ready_reason"] = status.ReadyReason.IN_PROGRESS
                status_update["ready_message"] = "Verify InProgress"

            status_update["updating_reason"] = status.UpdatingReason.VERIFY_WAIT
            status_update["updating_message"] = "Component verification incomplete"

        log.debug3("Updating status after reconcile: %s", status_update)
        self._update_resource_status(
            session.deploy_manager, session.cr_manifest, **status_update
        )

    def _update_error_status(
        self, resource: Union[dict, aconfig.Config], error: Exception
    ) -> dict:
        """Update the status of a resource after an error occurred. This function
        setups up it's own deploy manager and parses the resource. This way errors at any
        phase of reconciliation can still get updated

        Args:
            resource: Union[dict, aconfig.Config]
                The resource that's status is being updated
            error: Exception
                The exception that stopped the reconciliation

        Returns:
            status: dict
                The updated status after the error message
        """
        cr_manifest = self.parse_manifest(resource)
        deploy_manager = self.setup_deploy_manager(resource)

        # Get the completion state if possible
        component_state = getattr(error, "completion_state", None)

        # Expected Oper8 Errors
        if isinstance(error, PreconditionError):
            status_update = {
                "updating_reason": status.UpdatingReason.PRECONDITION_WAIT,
                "updating_message": str(error),
                "component_state": component_state,
            }
        elif isinstance(error, (VerificationError, Oper8ExpectedError)):
            status_update = {
                "updating_reason": status.UpdatingReason.VERIFY_WAIT,
                "updating_message": str(error),
                "component_state": component_state,
            }
        elif isinstance(error, ConfigError):
            status_update = {
                "ready_reason": status.ReadyReason.CONFIG_ERROR,
                "ready_message": str(error),
                "updating_reason": status.UpdatingReason.ERRORED,
                "updating_message": str(error),
                "component_state": component_state,
            }
        elif isinstance(error, ClusterError):
            status_update = {
                "updating_reason": status.UpdatingReason.CLUSTER_ERROR,
                "updating_message": str(error),
                "component_state": component_state,
            }

        elif isinstance(error, (RolloutError, Oper8FatalError)):
            status_update = {
                "ready_reason": status.ReadyReason.ERRORED,
                "ready_message": str(error),
                "updating_reason": status.UpdatingReason.ERRORED,
                "updating_message": str(error),
                "component_state": component_state,
            }

        # Catchall for non oper8 errors
        else:
            status_update = {
                "ready_reason": status.ReadyReason.ERRORED,
                "ready_message": str(error),
                "updating_reason": status.UpdatingReason.ERRORED,
                "updating_message": str(error),
            }

        return self._update_resource_status(
            deploy_manager, cr_manifest, **status_update
        )
